import logging
from abc import ABCMeta, abstractmethod
from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING

from lettrade.account import Account
from lettrade.commander import Commander
from lettrade.data import DataFeed, DataFeeder

from .execution import Execution
from .order import Order, OrderResult
from .position import Position

if TYPE_CHECKING:
    from lettrade.brain import Brain


logger = logging.getLogger(__name__)


class ExchangeState(int, Enum):
    Init = 1
    Start = 2
    Run = 3
    Stop = 4


class Exchange(metaclass=ABCMeta):
    """Base Exchange class to handle trading"""

    datas: list[DataFeed]
    """List of all available DataFeed"""
    data: DataFeed
    """main DataFeed"""

    executions: dict[str, Execution]
    """Execution dict by `Execution.id` key"""
    orders: dict[str, Order]
    """Available Order dict by `Order.id` key"""
    history_orders: dict[str, Order]
    """History Order dict by `Order.id` key"""
    positions: dict[str, Position]
    """Available Position dict by `Position.id` key"""
    history_positions: dict[str, Position]
    """History Position dict by `Position.id` key"""

    _config: dict

    _brain: "Brain"
    _feeder: DataFeeder
    _account: Account
    _commander: Commander

    _state: ExchangeState

    def __init__(self, **kwargs):
        self._config = kwargs

        self.orders = dict()
        self.history_orders = dict()
        self.positions = dict()
        self.history_positions = dict()
        self.executions = dict()

        self._state = ExchangeState.Init

    def init(
        self,
        brain: "Brain",
        feeder: DataFeeder,
        account: Account,
        commander: Commander,
    ) -> None:
        """Init Exchange dependencies.
        Set data/datas from feeder.
        Set state to `ExchangeState.Start`.

        Args:
            brain (Brain): _description_
            feeder (DataFeeder): _description_
            account (Account): _description_
            commander (Commander): _description_
        """
        self._brain = brain
        self._feeder = feeder
        self._account = account
        self._commander = commander

        self.data = self._feeder.data
        self.datas = self._feeder.datas

        self._account.init(exchange=self)

        self._state = ExchangeState.Start

    def start(self) -> None:
        """Start of Exchange by:
        - Start account.
        - Set state to `ExchangeState.Run` ready for next().
        """
        self._account.start()
        self._state = ExchangeState.Run

    def next(self):
        """Call after data feeded and before strategy.next()"""
        self._account.next()

    def next_next(self):
        """Call after strategy.next()"""
        self._account.next_next()

    def stop(self) -> None:
        """Stop Exchange"""
        self._state = ExchangeState.Stop
        self._account.stop()

    def on_execution(
        self,
        execution: Execution,
        broadcast: bool | None = True,
        **kwargs,
    ) -> None:
        """Receive Execution event from exchange then store and notify Brain

        Args:
            execution (Execution): _description_
            broadcast (bool | None, optional): _description_. Defaults to True.
        """
        self.on_executions(executions=[execution], broadcast=broadcast, **kwargs)

    def on_executions(
        self,
        executions: list[Execution],
        broadcast: bool | None = True,
        **kwargs,
    ) -> None:
        """
        Receive Execution event from exchange then store and notify Brain
        """
        for execution in executions:
            if not isinstance(execution, Execution):
                raise RuntimeError(f"{execution} is not instance of type Execution")

            if execution.id in self.executions:
                # Merge to keep Execution handler for strategy using
                # when strategy want to store Execution object
                # and object will be automatic update directly
                self.executions[execution.id].merge(execution)
                execution = self.executions[execution.id]
            else:
                self.executions[execution.id] = execution

        if self._state != ExchangeState.Run:
            return

        if broadcast:
            self._brain.on_executions(executions)

    def on_order(
        self,
        order: Order,
        broadcast: bool | None = True,
        **kwargs,
    ) -> None:
        """Receive order event from exchange then store and notify Brain

        Args:
            order (Order): _description_
            broadcast (bool | None, optional): _description_. Defaults to True.
        """
        self.on_orders(orders=[order], broadcast=broadcast, **kwargs)

    def on_orders(
        self,
        orders: list[Order],
        broadcast: bool | None = True,
        **kwargs,
    ) -> None:
        """Receive orders event from exchange then store and notify Brain

        Args:
            orders (list[Order]): _description_
            broadcast (bool | None, optional): _description_. Defaults to True.

        Raises:
            RuntimeError: _description_
        """
        for order in orders:
            if not isinstance(order, Order):
                raise RuntimeError(f"{order} is not instance of type Order")

            if order.is_closed:
                if order.id in self.history_orders:
                    logger.warning("Order closed recall: %s", order)

                self.history_orders[order.id] = order
                if order.id in self.orders:
                    del self.orders[order.id]
            else:
                if order.id in self.history_orders:
                    raise RuntimeError(f"Order {order.id} closed")

                if order.id in self.orders:
                    # Merge to keep Order handler for strategy using
                    # when strategy want to store Order object
                    # and object will be automatic update directly
                    self.orders[order.id].merge(order)
                    order = self.orders[order.id]
                else:
                    self.orders[order.id] = order

        if self._state != ExchangeState.Run:
            return

        if broadcast:
            self._brain.on_orders(orders)

    def on_position(
        self,
        position: Position,
        broadcast: bool | None = True,
        **kwargs,
    ) -> None:
        """Receive Position event from exchange then store and notify Brain

        Args:
            position (Position): new comming `Position`
            broadcast (bool | None, optional): Flag notify to Brain. Defaults to True.

        Raises:
            RuntimeError: validate `Position` instance
        """
        self.on_positions(positions=[position], broadcast=broadcast, **kwargs)

    def on_positions(
        self,
        positions: list[Position],
        broadcast: bool | None = True,
        **kwargs,
    ) -> None:
        """Receive Positions event from exchange then store and notify Brain

        Args:
            positions (list[Position]): list of new comming `Position`
            broadcast (bool | None, optional): Flag notify to Brain. Defaults to True.

        Raises:
            RuntimeError: validate `Position` instance
        """
        for position in positions:
            if not isinstance(position, Position):
                raise RuntimeError(f"{position} is not instance of type Position")

            if position.is_exited:
                if position.id in self.history_positions:
                    logger.warning("Position exited recall: %s", position)

                self.history_positions[position.id] = position
                if position.id in self.positions:
                    del self.positions[position.id]

                # self._account._on_position_exit(position)
            else:
                if position.id in self.history_positions:
                    raise RuntimeError(f"Position {position.id} closed: {position}")
                if position.id in self.positions:
                    # Merge to keep Position handler for strategy using
                    # when strategy want to store Position object
                    # and object will be automatic update directly
                    self.positions[position.id].merge(position)
                    position = self.positions[position.id]
                else:
                    self.positions[position.id] = position
                    # self._account._on_position_entry(position)

        self._account.on_positions(positions)

        if self._state != ExchangeState.Run:
            return

        if broadcast:
            self._brain.on_positions(positions)

    def on_notify(self, *args, **kwargs):
        return self._brain.on_notify(*args, **kwargs)

    @abstractmethod
    def new_order(
        self,
        size: float,
        limit: float | None = None,
        stop: float | None = None,
        sl: float | None = None,
        tp: float | None = None,
        tag: str | None = None,
        *args,
        **kwargs,
    ) -> OrderResult:
        """Place new `Order`

        Args:
            size (float): _description_
            limit (float | None, optional): _description_. Defaults to None.
            stop (float | None, optional): _description_. Defaults to None.
            sl (float | None, optional): _description_. Defaults to None.
            tp (float | None, optional): _description_. Defaults to None.
            tag (str | None, optional): _description_. Defaults to None.

        Returns:
            OrderResult: _description_
        """
        raise NotImplementedError("Exchange.new_order not implement yet")

    # Alias
    @property
    def now(self) -> datetime:
        return self.data.now



import logging
from typing import Any

import pandas as pd

from lettrade import (
    OrderResult,
    OrderResultError,
    OrderState,
    OrderType,
    PositionResult,
    PositionResultError,
    PositionState,
    TradeSide,
)
from lettrade.exchange.live import (
    LetLiveOrderInvalidException,
    LiveDataFeed,
    LiveExchange,
    LiveExecution,
    LiveOrder,
    LivePosition,
)

from .api import CCXTAPI

logger = logging.getLogger(__name__)


class CCXTExecution(LiveExecution):
    """Execution for CCXT"""

    @classmethod
    def from_raw(
        cls,
        raw,
        exchange: "LiveExchange",
        api: CCXTAPI = None,
    ) -> "CCXTExecution":
        """Building new CCXTExecution from live api raw object

        Args:
            raw (_type_): _description_
            exchange (LiveExchange): _description_

        Returns:
            CCXTExecution: _description_
        """

        return cls(
            exchange=exchange,
            id=raw.ticket,
            # TODO: Fix by get data from symbol
            data=exchange.data,
            # TODO: size and type from raw.type
            size=raw.volume,
            price=raw.price,
            # TODO: set bar time
            at=None,
            order_id=raw.order,
            position_id=raw.position_id,
            tag=raw.comment,
            api=api,
            raw=raw,
        )


class CCXTOrder(LiveOrder):
    """Order for CCXT"""

    exchange: "LiveExchange"

    def __init__(self, is_real: bool = True, **kwargs):
        super().__init__(**kwargs)

        self.is_real: bool = is_real
        """Flag to check `Order` is real, cannot duplicate id, cannot recall from history"""

    def place(self) -> OrderResult:
        """_summary_

        Raises:
            RuntimeError: _description_

        Returns:
            OrderResult: _description_
        """
        if self.state != OrderState.Pending:
            raise RuntimeError(f"Order {self.id} state {self.state} is not Pending")

        try:
            result = self._api.order_open(self)

            self.raw = result
            self.id = result.order

            # TODO: get current order time
            return super(LiveOrder, self).place(at=self.data.l.index[0], raw=result)
        except LetLiveOrderInvalidException as e:
            error = OrderResultError(
                error=e.message,
                order=self,
                raw=e.raw,
            )
            logger.error("Place order %s", str(error))
            self.exchange.on_notify(error=error)
            return error

    def update(
        self,
        limit_price: float | None = None,
        stop_price: float | None = None,
        sl: float | None = None,
        tp: float | None = None,
        caller: float | None = None,
        **kwargs,
    ) -> OrderResult:
        """_summary_

        Args:
            limit_price (float | None, optional): _description_. Defaults to None.
            stop_price (float | None, optional): _description_. Defaults to None.
            sl (float | None, optional): _description_. Defaults to None.
            tp (float | None, optional): _description_. Defaults to None.
            caller (float | None, optional): _description_. Defaults to None.

        Raises:
            RuntimeError: _description_

        Returns:
            OrderResult: _description_
        """
        if caller is self:
            raise RuntimeError(f"Order recusive update {self}")

        if self.parent is None:
            result = self._api.order_update(
                order=self,
                limit_price=limit_price,
                stop_price=stop_price,
                sl=sl,
                tp=tp,
                **kwargs,
            )
            return super(LiveOrder, self).update(
                limit_price=result.limit_price,
                stop_price=result.stop_price,
                sl=result.sl,
                tp=result.tp,
            )
        else:
            # SL/TP Order just a virtual order
            if caller is not self.parent:
                if self.is_sl_order:
                    self.parent.update(sl=stop_price, caller=self)
                elif self.is_tp_order:
                    self.parent.update(tp=limit_price, caller=self)
                else:
                    raise RuntimeError(f"Abandon order {self}")

            return super(LiveOrder, self).update(
                limit_price=limit_price,
                stop_price=stop_price,
            )

    def cancel(self, **kwargs) -> OrderResult:
        """Cancel order

        Returns:
            OrderResult: _description_
        """
        if not self.parent:
            # Abandon order
            result = self._api.order_close(order=self, **kwargs)
        else:
            # Virtual SL/TP order of trade
            result = None

        return super(LiveOrder, self).cancel(raw=result)

    @classmethod
    def from_raw(
        cls,
        raw: Any,
        exchange: "LiveExchange",
        api: CCXTAPI | None = None,
    ) -> "CCXTOrder | None":
        """_summary_

        Args:
            raw (Any): _description_.
            exchange (LiveExchange): _description_.
            api (CCXTAPI | None, optional): _description_. Defaults to None.

        Raises:
            NotImplementedError: _description_

        Returns:
            CCXTOrder: _description_
        """
        # DataFeed
        data = None
        for d in exchange.datas:
            if d.symbol == raw.symbol:
                data = d
                break
        if data is None:
            logger.warning("Raw order %s is not handling %s", raw.symbol, raw)
            return

        # Prices & Side & Type
        limit_price = None
        stop_price = None
        match raw.type:
            case MT5.ORDER_TYPE_BUY:
                side = TradeSide.Buy
                type = OrderType.Market
            case MT5.ORDER_TYPE_SELL:
                side = TradeSide.Sell
                type = OrderType.Market
            case MT5.ORDER_TYPE_BUY_LIMIT:
                side = TradeSide.Buy
                type = OrderType.Limit
                limit_price = raw.price_open
            case MT5.ORDER_TYPE_SELL_LIMIT:
                side = TradeSide.Sell
                type = OrderType.Limit
                limit_price = raw.price_open
            case MT5.ORDER_TYPE_BUY_STOP:
                side = TradeSide.Buy
                type = OrderType.Stop
                stop_price = raw.price_open
            case MT5.ORDER_TYPE_SELL_STOP:
                side = TradeSide.Sell
                type = OrderType.Stop
                stop_price = raw.price_open
            # case MT5.ORDER_TYPE_BUY_STOP_LIMIT:
            #     side = TradeSide.Buy
            #     type = OrderType.StopLimit
            #     # TODO
            #     limit_price = raw.price_open
            #     stop_price = raw.price_open
            # case MT5.ORDER_TYPE_SELL_STOP_LIMIT:
            #     side = TradeSide.Sell
            #     type = OrderType.StopLimit
            #     # TODO
            #     limit_price = raw.price_open
            #     stop_price = raw.price_open
            # case MT5.ORDER_TYPE_CLOSE_BY:
            case _:
                raise NotImplementedError(
                    f"Order type {raw.type} is not implement",
                    raw,
                )
        # State
        match raw.state:
            case MT5.ORDER_STATE_STARTED:
                state = OrderState.Pending
            case MT5.ORDER_STATE_PLACED:
                state = OrderState.Placed
            case MT5.ORDER_STATE_CANCELED:
                state = OrderState.Canceled
            case MT5.ORDER_STATE_PARTIAL:
                state = OrderState.Partial
            case MT5.ORDER_STATE_FILLED:
                state = OrderState.Filled
            case MT5.ORDER_STATE_REJECTED:
                state = OrderState.Canceled
            case MT5.ORDER_STATE_EXPIRED:
                state = OrderState.Canceled
            case MT5.ORDER_STATE_REQUEST_ADD:
                state = OrderState.Placed
            case MT5.ORDER_STATE_REQUEST_MODIFY:
                state = OrderState.Placed
            case MT5.ORDER_STATE_REQUEST_CANCEL:
                state = OrderState.Canceled
            case _:
                raise NotImplementedError(
                    f"Raw order state {raw.state} is not implement"
                )

        order = cls(
            exchange=exchange,
            id=raw.ticket,
            state=state,
            data=data,
            size=side * (raw.volume_current or raw.volume_initial),
            type=type,
            limit_price=limit_price,
            stop_price=stop_price,
            sl_price=raw.sl or None,
            tp_price=raw.tp or None,
            tag=raw.comment,
            placed_at=pd.to_datetime(raw.time_setup_msc, unit="ms", utc=True),
            api=api,
            raw=raw,
        )

        if hasattr(raw, "time_done_msc"):
            order.filled_price = raw.price_current
            order.filled_at = pd.to_datetime(raw.time_done_msc, unit="ms", utc=True)

        return order

    @classmethod
    def from_position(
        cls,
        position: "CCXTPosition",
        sl: float | None = None,
        tp: float | None = None,
    ) -> "CCXTOrder":
        """_summary_

        Args:
            position (CCXTPosition): _description_
            sl (float | None, optional): _description_. Defaults to None.
            tp (float | None, optional): _description_. Defaults to None.

        Raises:
            RuntimeError: _description_

        Returns:
            CCXTOrder: _description_
        """
        if not sl and not tp:
            raise RuntimeError("not sl and not tp")
        return cls(
            id=f"{position.id}-{'sl' if sl else 'tp'}",
            exchange=position.exchange,
            data=position.data,
            state=OrderState.Placed,
            type=OrderType.Stop if sl else OrderType.Limit,
            size=-position.size,
            limit_price=tp,
            stop_price=sl,
            parent=position,
            placed_at=position.entry_at,
            is_real=False,
        )


class CCXTPosition(LivePosition):
    """Position for CCXT"""

    exchange: "LiveExchange"

    def update(
        self,
        sl: float | None = None,
        tp: float | None = None,
        caller: float | None = None,
        **kwargs,
    ) -> PositionResult:
        """_summary_

        Args:
            sl (float | None, optional): _description_. Defaults to None.
            tp (float | None, optional): _description_. Defaults to None.
            caller (float | None, optional): _description_. Defaults to None.

        Raises:
            RuntimeError: _description_

        Returns:
            PositionResult: _description_
        """
        if not sl and not tp:
            raise RuntimeError("Update sl=None and tp=None")
        if caller is self:
            raise RuntimeError(f"Position recusive update {self}")

        result = self._api.position_update(position=self, sl=sl, tp=tp)
        if result.code != 0:
            logger.error("Update position %s", str(result))
            error = PositionResultError(
                error=result.error,
                position=self,
                raw=result,
            )
            self.exchange.on_notify(error=error)
            return error

        if sl is not None:
            if self.sl_order:
                if caller is not self.sl_order:
                    self.sl_order.update(stop_price=sl, caller=self)
            else:
                self.sl_order = self.exchange._order_cls.from_position(
                    position=self, sl=sl
                )

        if tp is not None:
            if self.tp_order:
                if caller is not self.tp_order:
                    self.tp_order.update(limit_price=tp, caller=self)
            else:
                self.tp_order = self.exchange._order_cls.from_position(
                    position=self, tp=tp
                )

        return super(LivePosition, self).update(raw=result)

    def exit(self) -> PositionResult:
        """_summary_

        Returns:
            PositionResult: _description_
        """
        result = self._api.position_close(position=self)
        if result.code != 0:
            logger.error("Update position %s", str(result))
            error = PositionResultError(
                error=result.error,
                position=self,
                raw=result,
            )
            self.exchange.on_notify(error=error)
            return error

        execution_raw = self._api.execution_get(id=result.execution_id)

        # TODO: execution object and event
        result.execution_raw = execution_raw

        return super(LivePosition, self).exit(
            price=result.price,
            at=pd.to_datetime(execution_raw.time_msc, unit="ms", utc=True),
            pl=execution_raw.profit,
            fee=execution_raw.fee + execution_raw.swap + execution_raw.commission,
            raw=result,
        )

    @classmethod
    def from_raw(
        cls,
        raw,
        exchange: "LiveExchange",
        state: PositionState = PositionState.Open,
        data: "LiveDataFeed" = None,
        api: CCXTAPI = None,
    ) -> "CCXTPosition":
        """_summary_

        Args:
            raw (_type_): _description_
            exchange (LiveExchange): _description_
            data (LiveDataFeed, optional): _description_. Defaults to None.
            api (CCXTAPI, optional): _description_. Defaults to None.

        Raises:
            NotImplementedError: _description_

        Returns:
            CCXTPosition: _description_
        """
        # DataFeed
        if data is None:
            for d in exchange.datas:
                if d.symbol == raw.symbol:
                    data = d
                    break
            if data is None:
                logger.warning("Raw position %s is not handling %s", raw.symbol, raw)
                return

        # Side
        match raw.type:
            case MT5.POSITION_TYPE_BUY:
                side = TradeSide.Buy
            case MT5.POSITION_TYPE_SELL:
                side = TradeSide.Sell
            case _:
                raise NotImplementedError(
                    f"Position type {raw.type} is not implement",
                    raw,
                )

        position = cls(
            exchange=exchange,
            id=raw.ticket,
            data=data,
            state=state,
            size=side * raw.volume,
            entry_price=raw.price_open,
            entry_fee=raw.swap,
            entry_at=pd.to_datetime(raw.time_msc, unit="ms", utc=True),
            parent=None,
            tag=raw.comment,
            api=api,
            raw=raw,
        )

        # SL
        if raw.sl > 0.0:
            position.sl_order = exchange._order_cls.from_position(
                position=position, sl=raw.sl
            )
            exchange.on_order(position.sl_order)

        # TP
        if raw.tp > 0.0:
            position.tp_order = exchange._order_cls.from_position(
                position=position, tp=raw.tp
            )
            exchange.on_order(position.tp_order)

        return position

    # @classmethod
    # def from_id(
    #     cls,
    #     id: str,
    #     exchange: "LiveExchange",
    #     data: LiveDataFeed = None,
    #     api: CCXTAPI = None,
    # ) -> "CCXTPosition":
    #     if api is None:
    #         api = exchange._api

    #     raws = api.positions_get(id=id)
    #     return cls.from_raw(
    #         raw=raws[0],
    #         exchange=exchange,
    #         data=data,
    #         api=api,
    #     )



import logging
from enum import Enum
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from lettrade import DataFeed, Exchange

logger = logging.getLogger(__name__)


class TradeSide(int, Enum):
    """Side of Trade"""

    Buy = 1
    """Buy/Long side"""
    Sell = -1
    """Sell/Short side"""

    def lower(self) -> str:
        return self.name.lower()

    def upper(self) -> str:
        return self.name.upper()


class OrderType(str, Enum):
    """Order type"""

    Market = "market"
    """Market order"""
    Limit = "limit"
    """Limit order"""
    Stop = "stop"
    """Stop order"""
    StopLimit = "stoplimit"
    """Stop-Limit order"""


class OrderState(str, Enum):
    """Order state"""

    Pending = "pending"
    """Pending order, wait for exchange accept"""
    Placed = "place"
    """Placed order on exchange"""
    Partial = "partial"
    """Partial filled, still wait for full fill"""
    Filled = "filled"
    """Full filled"""
    Canceled = "canceled"
    """Canceled"""


class PositionState(str, Enum):
    """Position state"""

    Open = "open"
    """Opening state"""
    Exit = "exit"
    """Exited state"""


class BaseTransaction:
    """Base class of Execution/Order/Position"""

    def __init__(
        self,
        id: str,
        exchange: "Exchange",
        data: "DataFeed",
        size: float,
    ) -> None:
        """_summary_

        Args:
            id (str): _description_
            exchange (Exchange): _description_
            data (DataFeed): _description_
            size (float): _description_
        """
        self.id: str = id
        self.exchange: "Exchange" = exchange
        self.data: "DataFeed" = data
        self.size: float = size

    @property
    def is_long(self) -> bool:
        """True if side is long (size is positive).

        Returns:
            bool: True/False
        """
        return self.size > 0

    @property
    def is_short(self) -> bool:
        """True if side is short (size is negative).

        Returns:
            bool: _description_
        """
        return self.size < 0

    @property
    def side(self) -> TradeSide:
        """True if side is short (size is negative).

        Returns:
            bool: _description_
        """
        return TradeSide.Sell if self.size < 0 else TradeSide.Buy


if __debug__:
    from lettrade.base.flag import validate_strategy_trade

    if validate_strategy_trade:

        import inspect

        def base_transaction__setattr__(self, name: str, value: Any) -> None:

            stack = inspect.stack()[1]
            caller_cls = stack[0].f_locals["self"].__class__

            from lettrade.strategy import Strategy

            if issubclass(caller_cls, Strategy):
                caller_method = stack[0].f_code.co_name
                logger.warning(
                    "Strategy %s.%s try to set %s=%s, run 'python -OO' to bypass warning",
                    caller_cls,
                    caller_method,
                    name,
                    value,
                )

            object.__setattr__(self, name, value)

        setattr(BaseTransaction, "__setattr__", base_transaction__setattr__)



"""Live trading base classes"""

from .account import LiveAccount
from .api import LiveAPI
from .data import LiveDataFeed
from .error import LetLiveAPIUnauthorizedException, LetLiveOrderInvalidException
from .exchange import LiveExchange
from .feeder import LiveDataFeeder
from .live import LetTradeLive, LetTradeLiveBot, let_live
from .trade import LiveExecution, LiveOrder, LivePosition



import pandas as pd


def ichimoku(
    dataframe: pd.DataFrame,
    conversion_line_window: int = 9,
    base_line_windows: int = 26,
    laggin_span: int = 52,
    displacement: int = 26,
    cloud: bool = False,
    prefix: str = "ichimoku_",
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_kwargs: dict | None = None,
) -> dict[str, pd.Series] | pd.DataFrame:
    """Ichimoku cloud indicator

    Note:
        Do not use `chikou_span` for backtesting.
        It looks into the future, is not printed by most charting platforms.
        It is only useful for visual analysis

    Args:
        dataframe (pd.DataFrame): Dataframe containing OHLCV data
        conversion_line_window (int, optional): Conversion line Window. Defaults to 9.
        base_line_windows (int, optional): Base line Windows. Defaults to 26.
        laggin_span (int, optional): Lagging span window. Defaults to 52.
        displacement (int, optional): Displacement (shift). Defaults to 26.
        cloud (bool, optional): Add cloud direction. Defaults to False.
        prefix (str, optional): _description_. Defaults to "ichimoku_".
        inplace (bool, optional): _description_. Defaults to False.
        plot (bool | list, optional): _description_. Defaults to False.
        plot_kwargs (dict | None, optional): _description_. Defaults to None.

    Returns:
        dict[str, pd.Series] | pd.DataFrame: {tenkan_sen, kijun_sen, senkou_span_a, senkou_span_b, leading_senkou_span_a, leading_senkou_span_b, chikou_span, cloud_white, cloud_black}
    """
    if __debug__:
        if not isinstance(dataframe, pd.DataFrame):
            raise RuntimeError(
                f"dataframe type '{type(dataframe)}' "
                "is not instance of pandas.DataFrame"
            )
        if plot and not inplace:
            raise RuntimeError("Cannot plot when inplace=False")

    tenkan_sen = (
        dataframe["high"].rolling(window=conversion_line_window).max()
        + dataframe["low"].rolling(window=conversion_line_window).min()
    ) / 2

    kijun_sen = (
        dataframe["high"].rolling(window=base_line_windows).max()
        + dataframe["low"].rolling(window=base_line_windows).min()
    ) / 2

    leading_senkou_span_a = (tenkan_sen + kijun_sen) / 2

    leading_senkou_span_b = (
        dataframe["high"].rolling(window=laggin_span).max()
        + dataframe["low"].rolling(window=laggin_span).min()
    ) / 2

    senkou_span_a = leading_senkou_span_a.shift(displacement - 1)

    senkou_span_b = leading_senkou_span_b.shift(displacement - 1)

    chikou_span = dataframe["close"].shift(-displacement + 1)

    # Result is inplace or new dict
    result = dataframe if inplace else {}

    result[f"{prefix}tenkan_sen"] = tenkan_sen
    result[f"{prefix}kijun_sen"] = kijun_sen
    result[f"{prefix}senkou_span_a"] = senkou_span_a
    result[f"{prefix}senkou_span_b"] = senkou_span_b
    result[f"{prefix}leading_senkou_span_a"] = leading_senkou_span_a
    result[f"{prefix}leading_senkou_span_b"] = leading_senkou_span_b
    result[f"{prefix}chikou_span"] = chikou_span

    if cloud:
        cloud_white = senkou_span_a > senkou_span_b
        cloud_black = senkou_span_b > senkou_span_a
        result[f"{prefix}cloud_white"] = cloud_white
        result[f"{prefix}cloud_black"] = cloud_black

    if plot:
        if plot_kwargs is None:
            plot_kwargs = dict()

        if isinstance(plot, list):
            if f"{prefix}tenkan_sen" in plot:
                plot_kwargs.update(tenkan_sen=f"{prefix}tenkan_sen")
            if f"{prefix}kijun_sen" in plot:
                plot_kwargs.update(kijun_sen=f"{prefix}kijun_sen")
            if f"{prefix}senkou_span_a" in plot:
                plot_kwargs.update(senkou_span_a=f"{prefix}senkou_span_a")
            if f"{prefix}senkou_span_b" in plot:
                plot_kwargs.update(senkou_span_b=f"{prefix}senkou_span_b")
            if f"{prefix}chikou_span" in plot:
                plot_kwargs.update(chikou_span=f"{prefix}chikou_span")
        else:
            plot_kwargs.update(
                tenkan_sen=f"{prefix}tenkan_sen",
                kijun_sen=f"{prefix}kijun_sen",
                senkou_span_a=f"{prefix}senkou_span_a",
                senkou_span_b=f"{prefix}senkou_span_b",
                chikou_span=f"{prefix}chikou_span",
            )
        from lettrade.indicator.plot import IndicatorPlotter
        from lettrade.plot.plotly import plot_ichimoku

        IndicatorPlotter(dataframe=dataframe, plotter=plot_ichimoku, **plot_kwargs)
    return result



import time
from datetime import datetime

from lettrade.data import DataFeeder, TimeFrame

from .api import LiveAPI
from .data import LiveDataFeed


class LiveDataFeeder(DataFeeder):
    # Class properties
    _api_cls: type[LiveAPI] = LiveAPI
    _data_cls: type[LiveDataFeed] = LiveDataFeed

    # Object properties
    datas: list[LiveDataFeed]
    data: LiveDataFeed

    _api: LiveAPI
    _tick: bool
    _start_size: int
    _config: dict
    _wait_timeframe: TimeFrame

    def __init__(
        self,
        api: LiveAPI | None = None,
        tick: bool = 5,
        start_size: int = 500,
        api_kwargs: dict | None = None,
        **kwargs,
    ) -> None:
        """
        tick:
            tick < 0: no tick, just get completed bar
            tick == 0: wait until new bar change value
            tick > 0: sleep tick time (in seconds) then update
        """
        super().__init__()

        # API init
        if api is None:
            if api_kwargs is None:
                raise RuntimeError("api or api_kwargs cannot missing")
            api = self._api_cls(**api_kwargs)

        self._api = api
        self._tick = tick
        self._start_size = start_size
        self._config = kwargs

        if isinstance(self._tick, int):
            self._wait_timeframe = TimeFrame(f"{self._tick}s")
        else:
            self._wait_timeframe = None

    def alive(self):
        return self._api.heartbeat()

    def start(self, size: int = 0):
        if not size:
            size = self._start_size

        for data in self.datas:
            data.next(size=size, tick=self._tick)

    def next(self):
        if self._tick > 0:
            time.sleep(self._wait_duration())
        for data in self.datas:
            data.next(tick=self._tick)

    def _wait_duration(self) -> float:
        now = datetime.now()
        delta = self._wait_timeframe.ceil(now) - now
        return delta.total_seconds()

    ### Extend
    def data_new(self, **kwargs):
        return self._data_cls(api=self._api, **kwargs)

    def market(self, symbol):
        return self._api.market(symbol)

    def markets(self, search=None):
        return self._api.markets(search=search)

    # @classmethod
    # def instance(
    #     cls,
    #     api: LiveAPI | None = None,
    #     api_kwargs: dict | None = None,
    #     **kwargs,
    # ) -> "LiveDataFeed":
    #     """_summary_

    #     Args:
    #         api (LiveAPI, optional): _description_. Defaults to None.
    #         api_kwargs (dict, optional): _description_. Defaults to None.

    #     Raises:
    #         RuntimeError: Missing api requirement

    #     Returns:
    #         LiveDataFeed: DataFeed object
    #     """
    #     if api is None:
    #         if api_kwargs is None:
    #             raise RuntimeError("api or api_kwargs cannot missing")
    #         api = cls._api_cls(**api_kwargs)
    #     obj = cls(api=api, **kwargs)
    #     return obj



from .rsi import rsi


def pandas_inject(obj: object | None = None):
    if obj is None:
        from pandas.core.base import PandasObject

        obj = PandasObject

    from ..series import series_indicator_inject

    obj.rsi = series_indicator_inject(rsi)



from typing import Literal

import pandas as pd
import talib.abstract as ta

from ..series import series_init
from ..utils import talib_ma_mode


def bollinger_bands(
    series: pd.Series | str = "close",
    window: int = 0,
    std: int | float = 0,
    ma_mode: Literal["ema", "sma"] = "sma",
    dataframe: pd.DataFrame = None,
    prefix: str = "bb_",
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_kwargs: dict | None = None,
    **kwargs,
) -> dict[str, pd.Series] | pd.DataFrame:
    """_summary_

    Args:
        series (pd.Series | str, optional): _description_. Defaults to "close".
        window (int, optional): _description_. Defaults to 0.
        std (int | float, optional): _description_. Defaults to 0.
        ma_mode (Literal[&quot;ema&quot;, &quot;sma&quot;], optional): _description_. Defaults to "sma".
        dataframe (pd.DataFrame, optional): _description_. Defaults to None.
        prefix (str, optional): _description_. Defaults to "bb_".
        inplace (bool, optional): _description_. Defaults to False.
        plot (bool | list, optional): Flag or list of columns name. Defaults to False.
        plot_kwargs (dict | None, optional): _description_. Defaults to None.

    Example:
        ```python
        df.i.bollinger_bands(window=20, std=2.0, inplace=True, plot=True)
        ```

    Raises:
        RuntimeError: _description_

    Returns:
        dict[str, pd.Series] | pd.DataFrame: _description_
    """
    if __debug__:
        if window <= 0:
            raise RuntimeError(f"Window {window} is invalid")
        if std <= 0:
            raise RuntimeError(f"Std {std} is invalid")
        if plot and not inplace:
            raise RuntimeError("Cannot plot when inplace=False")

    series = series_init(series=series, dataframe=dataframe, inplace=inplace)

    i_upper, i_basis, i_lower = ta.BBANDS(
        series,
        timeperiod=window,
        nbdevup=std,
        nbdevdn=std,
        matype=talib_ma_mode(ma_mode),
        **kwargs,
    )

    # Result is inplace or new dict
    result = dataframe if inplace else {}
    result[f"{prefix}upper"] = i_upper
    result[f"{prefix}basis"] = i_basis
    result[f"{prefix}lower"] = i_lower

    # Plot
    if plot:
        if plot_kwargs is None:
            plot_kwargs = dict()

        if isinstance(plot, list):
            if f"{prefix}upper" in plot:
                plot_kwargs.update(upper=f"{prefix}upper")
            if f"{prefix}basis" in plot:
                plot_kwargs.update(basis=f"{prefix}basis")
            if f"{prefix}lower" in plot:
                plot_kwargs.update(lower=f"{prefix}lower")
        else:
            plot_kwargs.update(
                upper=f"{prefix}upper",
                basis=f"{prefix}basis",
                lower=f"{prefix}lower",
            )

        from lettrade.indicator.plot import IndicatorPlotter
        from lettrade.plot.plotly import plot_bollinger_bands

        IndicatorPlotter(
            dataframe=dataframe,
            plotter=plot_bollinger_bands,
            **plot_kwargs,
        )

    return result



def is_notebook_session():
    if __debug__:
        try:
            shell = get_ipython().__class__.__name__
            if shell == "ZMQInteractiveShell":
                return True  # Jupyter notebook or qtconsole
            elif shell == "TerminalInteractiveShell":
                return False  # Terminal running IPython
            else:
                return False  # Other type (?)
        except NameError:
            return False



from typing import Literal

import pandas as pd
import talib.abstract as ta

from ..utils import talib_ma


def keltner_channel(
    dataframe: pd.DataFrame,
    ma: int = 20,
    ma_mode: Literal[
        "sma", "ema", "wma", "dema", "tema", "trima", "kama", "mama", "t3"
    ] = "ema",
    atr: int = 20,
    shift: float = 1.6,
    prefix: str = "kc_",
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_kwargs: dict | None = None,
) -> dict[str, pd.Series] | pd.DataFrame:
    """_summary_

    Args:
        dataframe (pd.DataFrame): _description_
        ma (int, optional): _description_. Defaults to 20.
        ma_mode (Literal[ "sma", "ema", "wma", "dema", "tema", "trima", "kama", "mama", "t3" ], optional): _description_. Defaults to "ema".
        atr (int, optional): _description_. Defaults to 20.
        shift (float, optional): _description_. Defaults to 1.6.
        prefix (str, optional): _description_. Defaults to "kc_".
        inplace (bool, optional): _description_. Defaults to False.
        plot (bool | list, optional): _description_. Defaults to False.
        plot_kwargs (dict | None, optional): _description_. Defaults to None.

    Example:
        ```python
        df.i.keltner_channel(ma=20, atr=20, shift=1.6, inplace=True, plot=True)
        ```

    Raises:
        RuntimeError: _description_

    Returns:
        dict[str, pd.Series] | pd.DataFrame: {kc_upper, kc_middle, kc_lower}
    """
    if __debug__:
        if not isinstance(dataframe, pd.DataFrame):
            raise RuntimeError(
                f"dataframe type '{type(dataframe)}' "
                "is not instance of pandas.DataFrame"
            )
        if plot and not inplace:
            raise RuntimeError("Cannot plot when inplace=False")

    ma_fn = talib_ma(ma_mode)
    i_basis = ma_fn(dataframe, timeperiod=ma)

    i_atr = ta.ATR(dataframe, timeperiod=atr)
    i_upper = i_basis + shift * i_atr
    i_lower = i_basis - shift * i_atr

    # Result is inplace or new dict
    result = dataframe if inplace else {}
    result[f"{prefix}upper"] = i_upper
    result[f"{prefix}basis"] = i_basis
    result[f"{prefix}lower"] = i_lower

    # Plot
    if plot:
        if plot_kwargs is None:
            plot_kwargs = dict()

        if isinstance(plot, list):
            if f"{prefix}upper" in plot:
                plot_kwargs.update(upper=f"{prefix}upper")
            if f"{prefix}basis" in plot:
                plot_kwargs.update(basis=f"{prefix}basis")
            if f"{prefix}lower" in plot:
                plot_kwargs.update(lower=f"{prefix}lower")
        else:
            plot_kwargs.update(
                upper=f"{prefix}upper",
                basis=f"{prefix}basis",
                lower=f"{prefix}lower",
            )

        from lettrade.indicator.plot import IndicatorPlotter
        from lettrade.plot.plotly import plot_keltner_channel

        IndicatorPlotter(
            dataframe=dataframe,
            plotter=plot_keltner_channel,
            **plot_kwargs,
        )

    return result



import numpy as np
import pandas as pd
import pandas.testing as pdtest
import pytest

from lettrade.indicator import (
    above,
    below,
    crossover,
    crossunder,
    rolling_above,
    rolling_below,
    rolling_max,
    rolling_mean,
    rolling_min,
    s_above,
    s_below,
)


@pytest.fixture
def series1():
    return pd.Series([1] * 5 + [2] * 5 + [1] * 5)


@pytest.fixture
def series2():
    return pd.Series([2] * 5 + [1] * 5 + [4] * 5)


### Test signal
def test_s_above(series1: pd.Series, series2: pd.Series):
    i = s_above(series1, series2)
    values = pd.Series([0] * 5 + [100] * 5 + [0] * 5)

    pdtest.assert_series_equal(i, values)


def test_s_below(series1: pd.Series, series2: pd.Series):
    i = s_below(series1, series2)
    values = pd.Series([-100] * 5 + [0] * 5 + [-100] * 5)

    pdtest.assert_series_equal(i, values)


### Test indicator above/below
def test_above(series1: pd.Series, series2: pd.Series):
    i = above(series1, series2)
    values = pd.Series([0] * 5 + [100] * 5 + [0] * 5)

    pdtest.assert_series_equal(i, values)


def test_below(series1: pd.Series, series2: pd.Series):
    i = below(series1, series2)
    values = pd.Series([-100] * 5 + [0] * 5 + [-100] * 5)

    pdtest.assert_series_equal(i, values)


### Test indicator rolling_above/rolling_below
def test_rolling_above(series1: pd.Series, series2: pd.Series):
    i = rolling_above(series1, series2, window=3)
    values = pd.Series([0] * 5 + [0, 0, 100, 100, 100] + [0] * 5)

    pdtest.assert_series_equal(i, values)


def test_rolling_below(series1: pd.Series, series2: pd.Series):
    i = rolling_below(series1, series2, window=3)
    values = pd.Series([0, 0, -100, -100, -100] + [0] * 5 + [0, 0, -100, -100, -100])

    pdtest.assert_series_equal(i, values)


### Test indicator crossover/crossunder
def test_crossover(series1: pd.Series, series2: pd.Series):
    i = crossover(series1, series2)
    values = pd.Series([0] * 5 + [100] + [0] * 9)

    pdtest.assert_series_equal(i, values)


def test_crossunder(series1: pd.Series, series2: pd.Series):
    i = crossunder(series1, series2)
    values = pd.Series([0] * 10 + [-100] + [0] * 4)

    pdtest.assert_series_equal(i, values)


### Test indicator rolling_min/rolling_max/rolling_mean
def test_rolling_min(series1: pd.Series):
    i = rolling_min(series1, window=3)
    values = pd.Series([np.nan, np.nan, 1, 1, 1] + [1, 1, 2, 2, 2] + [1] * 5)

    pdtest.assert_series_equal(i, values)


def test_rolling_max(series1: pd.Series):
    i = rolling_max(series1, window=3)
    values = pd.Series([np.nan, np.nan, 1, 1, 1] + [2] * 5 + [2, 2, 1, 1, 1])

    pdtest.assert_series_equal(i, values)


def test_rolling_mean(series1: pd.Series):
    i = rolling_mean(series1, window=3)
    values = pd.Series(
        [np.nan, np.nan, 1, 1, 1] + [4 / 3, 5 / 3, 2, 2, 2] + [5 / 3, 4 / 3, 1, 1, 1]
    )

    pdtest.assert_series_equal(i, values)



from abc import ABC, abstractmethod


class Plotter(ABC):
    """
    Base class help to plot strategy
    """

    @abstractmethod
    def load(self):
        """Load plot config from `Strategy.plot()` and setup candlestick/equity"""

    @abstractmethod
    def plot(self, **kwargs):
        """Plot `equity`, `orders`, and `positions` then show"""

    @abstractmethod
    def stop(self):
        """stop plotter"""



from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING

import pandas as pd

from .base import BaseTransaction, PositionState
from .order import Order

if TYPE_CHECKING:
    from lettrade.account import Account
    from lettrade.data import DataFeed
    from lettrade.exchange import Exchange


class Position(BaseTransaction, metaclass=ABCMeta):
    """
    When an `Order` is filled, it results in an active `Position`.
    Find active positions in `Strategy.positions` and closed, settled positions in `Strategy.closed_positions`.
    """

    def __init__(
        self,
        id: str,
        exchange: "Exchange",
        data: "DataFeed",
        size: float,
        parent: "Order",
        state: PositionState = PositionState.Open,
        entry_price: float | None = None,
        entry_fee: float = 0.0,
        entry_at: pd.Timestamp | None = None,
        sl_order: Order | None = None,
        tp_order: Order | None = None,
        tag: str | None = None,
        **kwargs,
    ):
        super().__init__(
            id=id,
            exchange=exchange,
            data=data,
            size=size,
            **kwargs,
        )
        self._account: "Account" = self.exchange._account

        self.state: PositionState = state
        self.parent: "Order" = parent
        self.tag: str | None = tag

        self.entry_price: float | None = entry_price
        self.entry_fee: float = entry_fee
        self.entry_at: pd.Timestamp | None = entry_at

        self.exit_price: float | None = None
        self.exit_fee: float = 0.0
        self.exit_at: pd.Timestamp | None = None
        self.exit_pl: float | None = None

        self.sl_order: Order | None = sl_order
        self.tp_order: Order | None = tp_order

    def _repr_params(self):
        params = (
            f"id='{self.id}'"
            f", state='{self.state}'"
            f", size={round(self.size, 5)}"
            f", entry_price={self.entry_price}"
            f", exit_fee={self.exit_fee}"
            f", entry_at={self.entry_at}"
        )

        if self.sl_order:
            params += f", sl_order={self.sl_order}"
        if self.tp_order:
            params += f", tp_order={self.tp_order}"

        if self.is_exited:
            params += (
                f", exit_price={self.exit_price}"
                f", exit_fee={self.exit_fee}"
                f", exit_at={self.exit_at}"
                f", exit_pl={round(self.exit_pl,5)}"
            )

        if self.tag:
            params += f", tag='{self.tag}'"

        return params

    def __repr__(self):
        return f"<{self.__class__.__name__} {self._repr_params()}>"

    def entry(
        self,
        price: float,
        at: pd.Timestamp,
        fee: float,
        raw: object | None = None,
    ) -> "PositionResult":
        self.entry_price = price
        self.entry_at = at
        self.entry_fee = fee
        self.state = PositionState.Open
        self.exchange.on_position(self)
        return PositionResultOk(position=self, raw=raw)

    def update(self, raw: object | None = None, **kwargs) -> "PositionResult":
        self.exchange.on_position(self)
        return PositionResultOk(position=self, raw=raw)

    def exit(
        self,
        price: float,
        at: pd.Timestamp,
        pl: float,
        fee: float,
        raw: object | None = None,
    ) -> "PositionResult":
        if self.state != PositionState.Open:
            return PositionResultError(
                f"Position state {self.state} != PositionState.Open",
                code=-1,
                position=self,
                raw=raw,
            )

        self.exit_price = price
        self.exit_at = at
        self.exit_pl = pl
        self.exit_fee = fee
        self.state = PositionState.Exit
        self.exchange.on_position(self)
        return PositionResultOk(position=self, raw=raw)

    def merge(self, other: "Position") -> bool:
        """Merge position from another position has same id

        Args:
            other (Position): _description_

        Raises:
            RuntimeError: _description_
        """
        if other is self:
            return False

        if self.id != other.id:
            raise RuntimeError(f"Merge difference id {self.id} != {other.id} order")

        self.size = other.size

        if other.sl_order is not None:
            self.sl_order = other.sl_order
            self.sl_order.parent = self
        elif self.sl_order is not None:
            self.sl_order.cancel()
            self.sl_order = None

        if other.tp_order is not None:
            self.tp_order = other.tp_order
            self.tp_order.parent = self
        elif self.tp_order is not None:
            self.tp_order.cancel()
            self.tp_order = None

        if other.entry_at:
            self.entry_at = other.entry_at
        if other.entry_price:
            self.entry_price = other.entry_price

        if other.exit_at:
            self.exit_at = other.exit_at
        if other.exit_price:
            self.exit_price = other.exit_price
        if other.exit_fee:
            self.exit_fee = other.exit_fee
        if other.exit_pl:
            self.exit_pl = other.exit_pl

        if other.parent:
            self.parent = other.parent

        return True

    # Extra properties
    @property
    def sl(self) -> float | None:
        if self.sl_order:
            return self.sl_order.stop_price
        return None

    @property
    def tp(self) -> float | None:
        if self.tp_order:
            return self.tp_order.limit_price
        return None

    @property
    def is_opening(self) -> bool:
        """Flag to check Position state.

        Returns:
            bool: True if the trade opening
        """
        return self.state == PositionState.Open

    @property
    def is_exited(self) -> bool:
        """Flag to check Position state.

        Returns:
            bool: True if the trade exited
        """
        return self.state == PositionState.Exit

    @property
    def pl(self) -> float:
        """Estimate Profit or Loss of Position

        Returns:
            float: PnL
        """
        if self.state == PositionState.Exit:
            pl = self.exit_pl
        else:
            pl = self._account.pl(size=self.size, entry_price=self.entry_price)
        return pl + self.fee

    @property
    def fee(self) -> float:
        """Fee/Estimate Fee for trade

        Returns:
            float: Fee
        """
        return self.entry_fee + self.exit_fee


class PositionResult:
    """Result of `Position`"""

    def __init__(
        self,
        ok: bool = True,
        position: "Position | None" = None,
        raw: object | None = None,
    ) -> None:
        """_summary_

        Args:
            ok (bool | None, optional): Flag to check `Position` is success or not. Defaults to True.
            position (Position | None, optional): Position own the result. Defaults to None.
            raw (object | None, optional): Raw object of `Position`. Defaults to None.
        """
        self.ok: bool = ok
        self.position: "Position | None" = position
        self.raw: object | None = raw

    def _repr_params(self):
        params = f"ok={self.ok} position={self.position}"
        if self.raw is not None:
            params += f"raw='{self.raw}'"
        return params

    def __repr__(self):
        return f"<{self.__class__.__name__} {self._repr_params()}>"


class PositionResultOk(PositionResult):
    """Result of a success `Position`"""

    def __init__(
        self,
        position: "Position | None" = None,
        raw: object | None = None,
    ) -> None:
        """_summary_

        Args:
            position (Position | None, optional): Position own the result. Defaults to None.
            raw (object | None, optional): Raw object of `Position`. Defaults to None.
        """
        super().__init__(ok=True, position=position, raw=raw)


class PositionResultError(PositionResult):
    """Result of a error `Position`"""

    def __init__(
        self,
        error: str,
        position: "Position | None" = None,
        raw: object | None = None,
    ) -> None:
        """_summary_

        Args:
            error (str): Error message
            position (Position | None, optional): Position own the result. Defaults to None.
            raw (object | None, optional): Raw object of `Position`. Defaults to None.
        """
        super().__init__(ok=False, position=position, raw=raw)
        self.error: str = error

    def _repr_params(self):
        return f"error={self.error} {super()._repr_params()}"



import logging
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING

from lettrade.exchange import (
    Execution,
    Order,
    OrderResult,
    OrderState,
    OrderType,
    Position,
    PositionResult,
    PositionState,
)

from .api import LiveAPI

if TYPE_CHECKING:
    from .data import LiveDataFeed
    from .exchange import LiveExchange

logger = logging.getLogger(__name__)


class _LiveTrade:
    def __init__(
        self,
        exchange: "LiveExchange",
        api: LiveAPI | None = None,
        raw: object | None = None,
        **kwargs,
    ) -> None:
        super().__init__(exchange=exchange, **kwargs)
        self.raw: object = raw
        self._api: LiveAPI = api or exchange._api


class LiveExecution(_LiveTrade, Execution, metaclass=ABCMeta):
    """
    Execution for Live
    """

    def __init__(
        self,
        id: str,
        exchange: "LiveExchange",
        data: "LiveDataFeed",
        size: float,
        price: float,
        at: float,
        order_id: str | None = None,
        order: "Order | None" = None,
        position_id: str | None = None,
        position: "Position | None" = None,
        # tag: str | None = None,
        api: LiveAPI | None = None,
        raw: object | None = None,
        **kwargs,
    ):
        super().__init__(
            id=id,
            exchange=exchange,
            data=data,
            size=size,
            price=price,
            at=at,
            order_id=order_id,
            order=order,
            position_id=position_id,
            position=position,
            api=api,
            raw=raw,
            **kwargs,
        )
        # self.tag: str | None = tag

    @classmethod
    @abstractmethod
    def from_raw(cls, raw, exchange: "LiveExchange") -> "LiveExecution":
        """Building new LiveExecution from live api raw object

        Args:
            raw (_type_): _description_
            exchange (LiveExchange): _description_

        Returns:
            LiveExecution: _description_
        """
        raise NotImplementedError(cls)


class LiveOrder(_LiveTrade, Order, metaclass=ABCMeta):
    def __init__(
        self,
        id: str,
        exchange: "LiveExchange",
        data: "LiveDataFeed",
        size: float,
        state: OrderState = OrderState.Pending,
        type: OrderType = OrderType.Market,
        limit_price: float | None = None,
        stop_price: float | None = None,
        sl_price: float | None = None,
        tp_price: float | None = None,
        parent: "Position | None" = None,
        tag: str | None = None,
        api: LiveAPI | None = None,
        raw: object | None = None,
        **kwargs,
    ):
        super().__init__(
            id=id,
            exchange=exchange,
            data=data,
            size=size,
            state=state,
            type=type,
            limit_price=limit_price,
            stop_price=stop_price,
            sl_price=sl_price,
            tp_price=tp_price,
            parent=parent,
            tag=tag,
            api=api,
            raw=raw,
            **kwargs,
        )

    @abstractmethod
    def place(self) -> OrderResult:
        raise NotImplementedError(type(self))

        # if self.state != OrderState.Pending:
        #     raise RuntimeError(f"Order {self.id} state {self.state} is not Pending")

        # result = self._api.order_open(self)
        # self.raw = result
        # if result.code != 0:
        #     logger.error("Place order %s", str(result))
        #     error = OrderResultError(
        #         error=result.error,
        #         code=result.code,
        #         order=self,
        #         raw=result,
        #     )
        #     self.exchange.on_notify(error=error)
        #     return error

        # self.id = result.order
        # # TODO: get current order time
        # return super().place(at=self.data.l.index[0], raw=result)

    @abstractmethod
    def update(
        self,
        limit_price: float | None = None,
        stop_price: float | None = None,
        sl: float | None = None,
        tp: float | None = None,
        caller: float | None = None,
        **kwargs,
    ) -> OrderResult:
        raise NotImplementedError(type(self))
        # if caller is self:
        #     raise RuntimeError(f"Order recusive update {self}")

        # result = self._api.order_update(
        #     order=self,
        #     limit_price=limit_price,
        #     stop_price=stop_price,
        #     sl=sl,
        #     tp=tp,
        #     **kwargs,
        # )

        # return super().update(
        #     limit_price=result.limit_price,
        #     stop_price=result.stop_price,
        #     sl=result.sl,
        #     tp=result.tp,
        # )

    def cancel(self, **kwargs) -> OrderResult:
        """Cancel order

        Returns:
            OrderResult: _description_
        """
        result = self._api.order_close(order=self, **kwargs)
        return super().cancel(raw=result)

    @classmethod
    @abstractmethod
    def from_raw(cls, raw, exchange: "LiveExchange") -> "LiveOrder":
        """_summary_

        Args:
            raw (_type_): _description_
            exchange (LiveExchange): _description_

        Returns:
            LiveOrder: _description_
        """
        raise NotImplementedError(cls)

    @classmethod
    @abstractmethod
    def from_position(cls, position: "LivePosition", sl=None, tp=None) -> "LiveOrder":
        """_summary_

        Args:
            position (LivePosition): _description_
            sl (_type_, optional): _description_. Defaults to None.
            tp (_type_, optional): _description_. Defaults to None.

        Returns:
            LiveOrder: _description_
        """
        raise NotImplementedError(cls)


class LivePosition(_LiveTrade, Position, metaclass=ABCMeta):
    def __init__(
        self,
        id: str,
        exchange: "LiveExchange",
        data: "LiveDataFeed",
        size: float,
        parent: Order,
        tag: str | None = None,
        state: PositionState = PositionState.Open,
        entry_price: float | None = None,
        entry_fee: float = 0.0,
        entry_at: int | None = None,
        sl_order: Order | None = None,
        tp_order: Order | None = None,
        api: LiveAPI | None = None,
        raw: object | None = None,
        **kwargs,
    ):
        super().__init__(
            id=id,
            exchange=exchange,
            data=data,
            size=size,
            parent=parent,
            tag=tag,
            state=state,
            entry_price=entry_price,
            entry_fee=entry_fee,
            entry_at=entry_at,
            sl_order=sl_order,
            tp_order=tp_order,
            api=api,
            raw=raw,
            **kwargs,
        )

    @abstractmethod
    def update(
        self,
        sl: float | None = None,
        tp: float | None = None,
        caller: float | None = None,
        **kwargs,
    ) -> PositionResult:
        """_summary_

        Args:
            sl (float | None, optional): _description_. Defaults to None.
            tp (float | None, optional): _description_. Defaults to None.
            caller (float | None, optional): _description_. Defaults to None.

        Raises:
            NotImplementedError: _description_
        """
        raise NotImplementedError(type(self))

    @abstractmethod
    def exit(self) -> PositionResult:
        """_summary_

        Returns:
            bool: _description_
        """
        raise NotImplementedError(type(self))

    def merge(self, other: "LivePosition") -> bool:
        """Merge LivePosition from another

        Args:
            other (LivePosition): _description_

        Returns:
            bool: _description_
        """
        if not super().merge(other):
            return False
        self.raw = other.raw
        return True

    @classmethod
    @abstractmethod
    def from_raw(
        cls,
        raw,
        exchange: "LiveExchange",
        state: PositionState = PositionState.Open,
        **kwargs,
    ) -> "LivePosition":
        """_summary_

        Args:
            raw (_type_): _description_
            exchange (LiveExchange): _description_

        Returns:
            LivePosition: _description_
        """
        raise NotImplementedError(cls)

    # @classmethod
    # # @abstractmethod
    # def from_id(cls, id: str, exchange: "LiveExchange") -> "LivePosition":
    #     """_summary_

    #     Args:
    #         id (str): _description_
    #         exchange (LiveExchange): _description_

    #     Returns:
    #         LivePosition: _description_
    #     """
    #     raise NotImplementedError



import logging

from lettrade.data import DataFeeder, LetNoMoreDataFeedException

from .data import BackTestDataFeed

logger = logging.getLogger(__name__)


class BackTestDataFeeder(DataFeeder):
    """BackTest DataFeeder"""

    datas: list[BackTestDataFeed]
    data: BackTestDataFeed

    _start_size: int

    def __init__(self, start_size: int = 500) -> None:
        super().__init__()
        self._start_size = start_size

    @property
    def is_continous(self):
        """Flag check is realtime continous datafeeder"""
        return False

    def start(self, size: int = 0):
        # self._cleanup_data()
        if not size:
            size = self._start_size

        next = self.data.l.index[size]
        for data in self.datas:
            data.next(
                size=size,
                next=next,
                missing="bypass",
            )

    # def _cleanup_data(self):
    #     # Synchronize start time
    #     start = self.data.now
    #     tf = self.data.timeframe
    #     for data in self.datas:
    #         if data is self.data:
    #             continue

    #         if data.timeframe <= tf:
    #             data_start = start
    #         else:
    #             data_start = start - data.timeframe

    #         df = data[data["datetime"] < data_start]
    #         logger.info(
    #             "BackTestDataFeed %s dropped %d rows from start",
    #             data.name,
    #             len(df.index),
    #         )
    #         data.drop(index=df.index, inplace=True)
    #         data.reset_index(inplace=True)

    def next(self):
        next = self.data.l.index[1]

        no_next = 0
        for data in self.datas:
            if not data.next(next=next):
                if data is self.data:
                    raise LetNoMoreDataFeedException()
                no_next += 1

        if no_next >= len(self.datas):
            # Skip lastest available bar, because if next some data feeded some are not
            raise LetNoMoreDataFeedException()

    def alive(self):
        return self.data.alive()



from typing import TYPE_CHECKING

import pandas as pd

from lettrade.account import Account
from lettrade.data import DataFeed, DataFeeder
from lettrade.exchange import Exchange
from lettrade.strategy import Strategy

from .plot import Plotter

if TYPE_CHECKING:
    from lettrade.bot import LetTradeBot

DATAFRAME_PLOTTERS_NAME = "_lt_plotters"


class BotPlotter(Plotter):
    """
    Class help to plot bot result
    """

    bot: "LetTradeBot"
    feeder: DataFeeder
    exchange: Exchange
    account: Account
    strategy: Strategy

    datas: list[DataFeed]
    """All plotting datafeeds"""

    _datas_stored: list[DataFeed] | None = None

    def __init__(self, bot: "LetTradeBot") -> None:
        self.bot = bot
        self.feeder = bot.feeder
        self.exchange = bot.exchange
        self.account = bot.account
        self.strategy = bot.strategy

        self.datas = self.feeder.datas

    @property
    def data(self) -> DataFeed:
        """Get plotting main datafeed

        Returns:
            DataFeed: _description_
        """
        return self.datas[0]

    @data.setter
    def data(self, value: DataFeed) -> None:
        self.datas[0] = value

    @property
    def _data_stored(self) -> DataFeed:
        return self._datas_stored[0]

    def jump(
        self,
        since: int | str | pd.Timestamp | None = None,
        order_id: str | None = None,
        position_id: str | None = None,
        range: int = 300,
        name: str | None = None,
    ):
        """Jump to place on datefeed

        Args:
            since (int | str | pd.Timestamp | None, optional): Jump to index/datetime. Defaults to None.
            order_id (str | None, optional): Jump to order id. Defaults to None.
            position_id (str | None, optional): Jump to position id. Defaults to None.
            range (int, optional): number of candle plot. Defaults to 300.
            name (str | None, optional): _description_. Defaults to None.

        Raises:
            RuntimeError: _description_
        """
        if self._datas_stored is None:
            self._datas_stored = self.datas.copy()

        if since is None:
            if order_id is not None:  # Jump to order id
                if not isinstance(order_id, str):
                    order_id = str(order_id)

                if order_id in self.exchange.orders:
                    order = self.exchange.orders[order_id]
                elif order_id in self.exchange.history_orders:
                    order = self.exchange.history_orders[order_id]
                else:
                    raise RuntimeError(f"Order id {order_id} not found")

                loc = self._data_stored.l.index.get_loc(order.placed_at)
                since = loc - int(range / 2)

            elif position_id is not None:  # Jump to position id
                if not isinstance(position_id, str):
                    position_id = str(position_id)

                if position_id in self.exchange.positions:
                    position = self.exchange.positions[position_id]
                elif position_id in self.exchange.history_positions:
                    position = self.exchange.history_positions[position_id]
                else:
                    raise RuntimeError(f"Position id {position_id} not found")

                loc = self._data_stored.l.index.get_loc(position.entry_at)
                since = loc - int(range / 2)
            else:  # Reset
                self.jump_reset()
                return

        elif isinstance(since, str):  # Parse string to pd.Timestamp, then since=index
            since = pd.to_datetime(since, utc=True)
            since = self._data_stored.l.index.get_loc(since)
        elif isinstance(since, pd.Timestamp):  # Get index of Timestamp
            since = self._data_stored.l.index.get_loc(since)

        if name is None:
            name = self._data_stored.name

        # Since min at pointer_start
        if since < self._data_stored.l.pointer_start:
            since = self._data_stored.l.pointer_start
        # Since max at pointer_stop
        if since > self._data_stored.l.pointer_stop - range:
            since = self._data_stored.l.pointer_stop - range

        # Jump
        jump_start_dt = None
        jump_stop_dt = None
        for i, data in enumerate(self._datas_stored):
            if i == 0:
                self.datas[i] = data.__class__(
                    data=data.l[since : since + range],
                    name=data.name,
                    timeframe=data.timeframe,
                )
                jump_start_dt = self.data.index[0]
                jump_stop_dt = self.data.index[-1]
            else:
                self.datas[i] = data.__class__(
                    data=data.loc[
                        (data.index >= jump_start_dt) & (data.index <= jump_stop_dt)
                    ],
                    name=data.name,
                    timeframe=data.timeframe,
                )

            if hasattr(data, DATAFRAME_PLOTTERS_NAME):
                object.__setattr__(
                    self.datas[i],
                    DATAFRAME_PLOTTERS_NAME,
                    getattr(data, DATAFRAME_PLOTTERS_NAME),
                )

        # Reload data
        self.load()

    def jump_reset(self) -> bool:
        """Reset jump datafeeds back to bot datafeeds"""
        if not self._datas_stored or self.data is self._data_stored:
            return False

        self.datas = self._datas_stored.copy()
        return True



import logging
from datetime import timezone
from pathlib import Path

import pandas as pd
from pandas.core.base import PandasObject

logger = logging.getLogger(__name__)


def csv_export(
    dataframe: pd.DataFrame,
    path: str | Path = "data/data.csv",
    tz: timezone = None,
    round: int = 5,
    **kwargs,
) -> pd.DataFrame:
    """Dump DataFeed to csv file. Inject function `pandas.DataFrame.let_to_csv()`

    Args:
        dataframe (pd.DataFrame): _description_
        path (str | Path, optional): _description_. Defaults to "data/data.csv".
        tz (timezone, optional): _description_. Defaults to None.
        round (int, optional): _description_. Defaults to 5.

    Returns:
        pd.DataFrame: _description_
    """
    dataframe = dataframe.astype(
        dtype={
            "open": "float",
            "high": "float",
            "low": "float",
            "close": "float",
            "volume": "float",
        }
    )

    if not isinstance(dataframe.index, pd.DatetimeIndex):
        dataframe = dataframe.set_index("datetime")

    if tz is not None:
        dataframe.index = dataframe.index.tz_convert(tz)

    if round > 0:
        dataframe = dataframe.round(round)

    if not isinstance(path, Path):
        path = Path(path)

    path.parent.mkdir(parents=True, exist_ok=True)
    dataframe.to_csv(path, **kwargs)

    logger.info("Saved data to %s", path)
    return dataframe


PandasObject.let_to_csv = csv_export



import unittest

import numpy as np
import pandas as pd
import pandas_ta as pdta
import talib.abstract as ta
from pandas import testing as pdtest

from lettrade.exchange.backtest.data import CSVBackTestDataFeed


class IndicatorTestCase(unittest.TestCase):
    def setUp(self):
        self.data = CSVBackTestDataFeed("test/assets/EURUSD_1h-0_1000.csv")
        self.data._set_main()

        self.raw_data = pd.read_csv(
            "test/assets/EURUSD_1h-0_1000.csv",
            index_col=0,
            parse_dates=["datetime"],
        )

        # Ta-Lib
        self.ta_ema = ta.EMA(self.data, timeperiod=21)
        self.ta_ema_raw = ta.EMA(self.raw_data, timeperiod=21)

        # Pandas_TA
        self.pdta_ema = pdta.ema(self.data.close, length=21)
        self.pdta_ema_raw = pdta.ema(self.raw_data.close, length=21)

    def test_types(self):
        self.assertIsInstance(
            self.ta_ema,
            pd.Series,
            "TaLib EMA is not instance of pandas.Series",
        )
        self.assertIsInstance(
            self.pdta_ema,
            pd.Series,
            "Pandas_ta EMA is not instance of pandas.Series",
        )

    def test_indicator(self):
        pdtest.assert_series_equal(
            self.ta_ema,
            self.pdta_ema,
            check_names=False,
        )
        pdtest.assert_series_equal(
            self.ta_ema,
            self.ta_ema_raw,
            check_names=False,
            check_index_type=False,
        )
        pdtest.assert_series_equal(
            self.pdta_ema,
            self.pdta_ema_raw,
            check_names=False,
            check_index_type=False,
        )
        pdtest.assert_series_equal(
            self.ta_ema_raw,
            self.pdta_ema_raw,
            check_names=False,
        )

    def test_next(self):
        df = self.data.copy(deep=True)
        df["ema"] = self.ta_ema

        for i in range(0, len(df)):
            ema_value = df.l.ema[0]
            raw_ema_value = self.ta_ema_raw.iloc[i]

            # Next first to by pass next continue condiction
            df.next()

            if np.isnan(ema_value) and np.isnan(raw_ema_value):
                continue

            self.assertEqual(ema_value, raw_ema_value, f"EMA[{i}] is wrong")


if __name__ == "__main__":
    unittest.main(verbosity=2)



from abc import ABCMeta, abstractmethod
from enum import Enum
from typing import TYPE_CHECKING

from .error import LetAccountInsufficientException

if TYPE_CHECKING:
    from lettrade.exchange import Exchange, Position, TradeSide


class AccountType(str, Enum):
    """"""

    Hedging = "hedging"
    """"""
    Speculation = "speculation"
    """"""


class Account(metaclass=ABCMeta):
    """Manage account balance, leverage, commission. Risk calculate and control"""

    _exchange: "Exchange"
    _config: dict
    _type: str

    _risk: float
    _balance: float
    _margin: float
    _leverage: float
    _equities: dict[str, float]

    _do_equity_snapshot: bool

    def __init__(
        self,
        risk: float = 0.02,
        balance: float = 10_000,
        margin: float = 1.0,
        leverage: float = 1.0,
        type: AccountType = AccountType.Hedging,
        **kwargs,
    ) -> None:
        """_summary_

        Args:
            risk (float, optional): _description_. Defaults to 0.02.
            balance (float, optional): _description_. Defaults to 10_000.
            margin (float, optional): _description_. Defaults to 1.0.
            leverage (float, optional): _description_. Defaults to 1.0.
            **kwargs (dict, optional): Config of account. Defaults to {}.
        """
        self._risk = risk
        self._balance = balance
        self._margin = margin
        self._leverage = leverage
        self._type = type
        self._config = kwargs

        self._equities = dict()

        self._do_equity_snapshot = True  # Snapshot balance

    def init(self, exchange: "Exchange"):
        self._exchange = exchange

    def __repr__(self):
        return "<Account " + str(self) + ">"

    def start(self):
        """Start account"""

    def next(self):
        """Next account"""

    def next_next(self):
        """Call after strategy.next()"""
        self._equity_snapshot()

    def stop(self):
        """Stop account"""
        try:
            self._equity_snapshot()
        except LetAccountInsufficientException:
            pass

    def risk(self, side: "TradeSide", size: float, **kwargs) -> float:
        """Risk calculation"""
        if size is None:
            return side * abs(self._risk)
        return side * abs(size)

    def pl(self, size, entry_price: float, exit_price=None) -> float:
        """Estimate temporary profit and loss"""
        if exit_price is None:
            exit_price = self._exchange.data.l.open[0]

        return size * (exit_price - entry_price)

    @property
    def balance(self) -> float:
        """Balance of account

        Returns:
            float: _description_
        """
        return self._balance

    @property
    @abstractmethod
    def equity(self) -> float:
        """Equity value of account

        Returns:
            float: _description_
        """
        raise NotImplementedError(type(self))

    def _equity_snapshot(self):
        if self._do_equity_snapshot or len(self._exchange.positions) > 0:
            equity = self.equity

            bar = self._exchange.data.bar()
            self._equities[bar] = equity

            if equity <= 0:
                raise LetAccountInsufficientException()

            if self._do_equity_snapshot:
                self._do_equity_snapshot = False

    def on_positions(self, positions: list["Position"]):
        """Event positions updated"""
        if not self._do_equity_snapshot:
            self._do_equity_snapshot = True



from typing import Any

import pandas as pd
from pandas.core.indexing import _iLocIndexer

__LET_WRAPPER_KEY__ = "_lt_wrapper"


class LetILocWrapper:
    """Wrap iloc object from DataFeed"""

    _data: pd.DatetimeIndex
    _owner: "LetDataFeedWrapper"

    def __init__(self, data, owner) -> None:
        """_summary_

        Args:
            data (_type_): DataFeed or pd.Series
            owner (_type_): DataFeed object
        """
        self._data = data
        self._owner = owner
        # print("-----> ILocWrapper", data.name, type(data))

    # def __getattr__(self, name: str):
    #     if hasattr(self._data.iloc, name):
    #         return getattr(self._data.iloc, name)
    #     raise NotImplementedError(f"Method {name} is not implement yet")

    def __getitem__(self, item: int | slice):
        if isinstance(item, int):
            item += self._owner._pointer
        elif isinstance(item, slice):
            item = slice(
                item.start + self._owner._pointer,
                item.stop + self._owner._pointer,
                item.step,
            )
        else:
            raise NotImplementedError(
                f"Get item {item} type {type(item)} is not implement yet"
            )

        return self._data._values[item]

    # Property
    @property
    def pointer(self):
        return self._owner._pointer


class LetIndexWapper:
    """Wrap DataFeed.index"""

    _data: pd.DatetimeIndex
    _owner: "LetDataFeedWrapper"

    def __init__(self, data, owner) -> None:
        """_summary_

        Args:
            data (_type_): pandas.DatetimeIndex
            owner (_type_): DataFeed object
        """
        self._data = data
        self._owner = owner
        # print("-----> IndexWapper", data.name, type(data))

    def __getattr__(self, name: str):
        return getattr(self._data, name)
        # if hasattr(self._data, name):
        #     return getattr(self._data, name)
        # raise NotImplementedError(f"Method {name} is not implement yet")

    def __getitem__(self, item: int | slice):
        if isinstance(item, int):
            item += self._owner._pointer
        elif isinstance(item, slice):
            item = slice(
                item.start + self._owner._pointer,
                item.stop + self._owner._pointer,
                item.step,
            )
        else:
            raise NotImplementedError(
                f"Get item {item} type {type(item)} is not implement yet"
            )
        return self._data._values[item]

    def get_loc(self, dt: pd.Timestamp) -> int:
        loc = self._data.get_loc(dt)
        return loc - self._owner._pointer

    # Property
    @property
    def pointer(self):
        return self._owner._pointer


class LetSeriesWapper:
    """Wrap DataFeed column"""

    _data: pd.DatetimeIndex
    _owner: "LetDataFeedWrapper"
    _iloc: LetILocWrapper | _iLocIndexer

    def __init__(self, data, owner) -> None:
        self._data = data
        self._owner = owner
        self._iloc = LetILocWrapper(self._data, self._owner)

        # print("-----> SeriesWapper", data.name, type(data))

    def __getattr__(self, name: str):
        return getattr(self._data, name)
        # if hasattr(self._data, name):
        #     return getattr(self._data, name)
        # raise NotImplementedError(f"Method {name} is not implement yet")

    def __getitem__(self, item: int | slice | Any):
        return self._iloc[item]
        # if isinstance(item, (int, slice)):
        #     return self._iloc[item]
        # return self._data[item]

    # Property
    @property
    def iloc(self) -> LetILocWrapper | _iLocIndexer:
        return self._iloc

    @property
    def pointer(self):
        return self._owner._pointer


# @pd.api.extensions.register_dataframe_accessor("l")
class LetDataFeedWrapper:
    """Wrap DataFeed"""

    _data: pd.DataFrame
    _pointer: int
    _iloc: LetILocWrapper | _iLocIndexer

    def __init__(self, data: pd.DataFrame) -> None:
        """_summary_

        Args:
            data (pd.DataFrame): DataFeed object

        Raises:
            RuntimeError: _description_
        """
        # Validate new instance not load an existed wrapper
        if hasattr(data.index, __LET_WRAPPER_KEY__):
            raise RuntimeError("DataFeed.index reuses a loaded wrapper")
        for column in data.columns:
            if hasattr(data[column], __LET_WRAPPER_KEY__):
                raise RuntimeError(f"DataFeed.{column} reuses a loaded wrapper")

        self._pointer = 0
        self._data = data
        self._iloc = LetILocWrapper(self._data, self)

    def __getattr__(
        self, name: str
    ) -> pd.Series | pd.Index | pd.DatetimeIndex | LetSeriesWapper | LetIndexWapper:
        # if hasattr(self._data, name):
        result = getattr(self._data, name)

        # Exist
        if hasattr(result, __LET_WRAPPER_KEY__):
            return getattr(result, __LET_WRAPPER_KEY__)

        # Set wrapper
        if isinstance(result, pd.Series):
            wrapper = LetSeriesWapper(result, self)
            setattr(result, __LET_WRAPPER_KEY__, wrapper)
            return wrapper

        if isinstance(result, pd.DatetimeIndex):
            wrapper = LetIndexWapper(result, self)
            setattr(result, __LET_WRAPPER_KEY__, wrapper)
            return wrapper

        return result

        # raise NotImplementedError(f"Method {name} is not implement yet")

    def __getitem__(self, item: int | slice | str | Any) -> pd.DataFrame:
        # return self._iloc[item]
        # if isinstance(item, (int, slice)):
        #     return self._iloc[item]
        if isinstance(item, int):
            return self._data.iloc[item + self._pointer]
        elif isinstance(item, slice):
            item = slice(
                item.start + self._pointer,
                item.stop + self._pointer,
                item.step,
            )
            return self._data.iloc[item]
        if isinstance(item, str):
            return self.__getattr__(item)
        return self._data[item]

    # Function
    def next(self, size=1):
        """Move pointer to next"""
        self._pointer += size

    def go_start(self) -> None:
        """Move pointer to begin"""
        self._pointer = 0

    def go_stop(self) -> None:
        """Move pointer to end"""
        self._pointer = len(self._data) - 1

    def reset(self) -> None:
        """Reset pointer to begin"""
        self._pointer = 0

    # Property
    @property
    def iloc(self) -> LetILocWrapper | _iLocIndexer:
        """Get iloc wrapper of DataFeed"""
        return self._iloc

    @property
    def pointer(self):
        """Get current pointer value"""
        return self._pointer

    @property
    def pointer_start(self):
        """Get start pointer value"""
        return -self._pointer

    @property
    def pointer_stop(self):
        """Get stop pointer value"""
        return len(self._data) - self._pointer



import functools
import logging
import time
from datetime import datetime, timedelta
from subprocess import Popen
from typing import TYPE_CHECKING

from box import Box
from mt5linux import MetaTrader5 as MT5

from lettrade.exchange import OrderType
from lettrade.exchange.live import LetLiveOrderInvalidException, LiveAPI

if TYPE_CHECKING:
    from .metatrader import MetaTraderExchange
    from .trade import MetaTraderOrder, MetaTraderPosition

logger = logging.getLogger(__name__)


TIMEFRAME_L2M = {
    "1m": MT5.TIMEFRAME_M1,
    "2m": MT5.TIMEFRAME_M2,
    "3m": MT5.TIMEFRAME_M3,
    "4m": MT5.TIMEFRAME_M4,
    "5m": MT5.TIMEFRAME_M5,
    "6m": MT5.TIMEFRAME_M6,
    "10m": MT5.TIMEFRAME_M10,
    "12m": MT5.TIMEFRAME_M12,
    "15m": MT5.TIMEFRAME_M15,
    "20m": MT5.TIMEFRAME_M20,
    "30m": MT5.TIMEFRAME_M30,
    "1h": MT5.TIMEFRAME_H1,
    "2h": MT5.TIMEFRAME_H2,
    "3h": MT5.TIMEFRAME_H3,
    "4h": MT5.TIMEFRAME_H4,
    "6h": MT5.TIMEFRAME_H6,
    "8h": MT5.TIMEFRAME_H8,
    "12h": MT5.TIMEFRAME_H12,
    "1d": MT5.TIMEFRAME_D1,
    "1w": MT5.TIMEFRAME_W1,
    "1mn": MT5.TIMEFRAME_MN1,
}


class _RetryException(Exception):
    pass


def mt5_connection(api_function):
    @functools.wraps(api_function)
    def wrapper(self: "MetaTraderAPI", *args, api_retry: int = 3, **kwargs):
        while api_retry > 0:
            try:
                return api_function(self, *args, **kwargs)
            except _RetryException:
                pass

            logger.warning(
                "Retry to reconnect MetaTrader 5 RPC for functon %s", api_function
            )
            if not self._refresh_environments():
                return None

            time.sleep(1)
            api_retry -= 1

    return wrapper


class MetaTraderAPI(LiveAPI):
    """API to connect MetaTrader 5 Terminal"""

    _mt5: MT5
    _exchange: "MetaTraderExchange"
    _magic: int
    _config: dict

    _load_history_since: datetime
    _deal_time_checked: datetime
    _orders_stored: dict[int, object]
    _executions_stored: dict[int, object]
    _positions_stored: dict[int, object]

    _wine_process: Popen | None = None

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "_singleton"):
            cls._singleton = object.__new__(cls)
        return cls._singleton

    @classmethod
    def multiprocess(cls, kwargs, **other_kwargs):
        # Pop to remove parameter "wine", then subprocess will not start wine server
        wine = kwargs.pop("wine", None)
        if not wine:
            return

        cls._wine_run(wine)

    @classmethod
    def _wine_run(cls, wine: str):
        if cls._wine_process is not None and cls._wine_process.poll() is None:
            logger.info("Wine MetaTrader rpyc is running: %s", wine)
            return

        logger.info("Start new Wine MetaTrader rpyc from path: %s", wine)
        cls._wine_process = Popen(wine, shell=True)

        # Wait for wine start
        time.sleep(5)

    def __init__(
        self,
        login: int,
        password: str,
        server: str,
        timeout: int = 60,
        retry: int = 20,
        host: str = "localhost",
        port: int = 18812,
        wine: str | None = None,
        path: str | None = None,
        magic: int = 88888888,
        **kwargs,
    ):
        """_summary_

        Args:
            login (int): _description_
            password (str): _description_
            server (str): _description_
            timeout (int, optional): _description_. Defaults to 60.
            retry (int, optional): _description_. Defaults to 20.
            host (str, optional): _description_. Defaults to "localhost".
            port (int, optional): _description_. Defaults to 18812.
            wine (str | None, optional): _description_. Defaults to None.
            magic (int, optional): _description_. Defaults to 88888888.

        Raises:
            ConnectionRefusedError: _description_
            RuntimeError: _description_
        """
        # Parameters
        self._config = kwargs
        self._magic = magic

        self._load_history_since = datetime.now() - timedelta(days=7)
        self._deal_time_checked = datetime.now() - timedelta(days=1)
        self._orders_stored = dict()
        self._executions_stored = dict()
        self._positions_stored = dict()

        # Update config
        self._config.update(
            host=host,
            port=port,
            login=int(login),
            password=password,
            server=server,
            wine=wine,
            path=path,
            retry=retry,
        )

        # Start enviroments
        self._refresh_environments()

    def _refresh_environments(self) -> bool:
        # Check start wine server if not inited yet
        wine = self._config.get("wine", False)
        if wine:
            self.__class__._wine_run(wine)

        # Init MT5
        if not hasattr(self, "_mt5") or self._mt5.account_info() is None:
            host = self._config.get("host")
            port = self._config.get("port")

            try:
                self._mt5 = MT5(host=host, port=port)
            except ConnectionRefusedError as e:
                raise ConnectionRefusedError(
                    "Cannot connect to MetaTrader 5 Terminal rpyc server"
                ) from e
            except TimeoutError as e:
                raise RuntimeError("Timeout start MetaTrader 5 Terminal") from e

        # Login account
        account = self._mt5.account_info()
        login = self._config.get("login")
        if not account or account.login != login:
            kwargs = dict()
            kwargs["login"] = int(login)
            kwargs["password"] = self._config.get("password")
            kwargs["server"] = self._config.get("server")

            path = self._config.get("path", None)
            if path is not None:
                kwargs["path"] = path

            retry = self._config.get("retry")
            while retry > 0:
                login = self._mt5.initialize(**kwargs)
                if login:
                    break

                if __debug__:
                    logger.debug("Login retry: %d", retry)

                time.sleep(1)
                retry -= 1

            if retry == 0:
                raise RuntimeError(f"Cannot login {account}: {self._mt5.last_error()}")

            # Terminal
            terminal = self._mt5.terminal_info()
            logger.info("Terminal: %s", str(terminal))
            if not terminal.trade_allowed:
                logger.warning("Terminal trading mode is not allowed")

            # Account
            logger.info("Login success: %s %s", kwargs, self._mt5.version())

            # Preload trading data
            now = datetime.now()
            self._mt5.history_deals_get(self._deal_time_checked, now)
            self._mt5.history_orders_get(self._load_history_since, now)
            self._mt5.orders_get()
            self._mt5.positions_get()
            time.sleep(3)

        return True

    def start(self, exchange: "MetaTraderExchange"):
        """_summary_

        Args:
            exchange (MetaTraderExchange): _description_
        """
        self._exchange = exchange

        self._load_history_transactions()
        self._check_transaction_events()

    def stop(self):
        """Stop MetaTrader 5 Terminal"""
        self._mt5.shutdown()

    def next(self):
        """Next tick action"""
        self._check_transaction_events()

    def heartbeat(self) -> bool:
        """Heartbeat

        Returns:
            bool: _description_
        """
        return True

    # ----- Public
    # --- Market
    @mt5_connection
    def market(self, symbol: str, **kwargs) -> dict:
        """_summary_

        Args:
            symbol (str): _description_

        Returns:
            dict: _description_
        """
        raw = self._mt5.symbol_info(symbol)
        if raw is None:
            raise _RetryException()
        return self._market_parse_response(raw)

    @mt5_connection
    def markets(self, search: str | None = None, **kwargs) -> list[dict]:
        """The filter for arranging a group of necessary symbols.
        If the group is specified, the function returns only symbols meeting a specified criteria.

        Search example:
            Get symbols whose names do not contain USD, EUR, JPY and GBP
            `search="*,!*USD*,!*EUR*,!*JPY*,!*GBP*"`

        Args:
            search (str | None, optional): _description_. Defaults to None.

        Returns:
            list[dict]: _description_
        """
        raws = self._mt5.symbols_get(search)
        if raws is None:
            raise _RetryException()
        return [self._market_parse_response(raw) for raw in raws]

    def _market_parse_response(self, raw):
        raw = Box(dict(raw._asdict()))
        type = raw.path.split("\\", 1)[0]
        return Box(
            symbol=raw.name,
            type=type,
            description=raw.description,
            currency_base=raw.currency_base,
            currency_profit=raw.currency_profit,
            raw=raw,
        )

    # --- Tick
    @mt5_connection
    def tick_get(self, symbol: str, **kwargs) -> dict:
        """_summary_

        Args:
            symbol (str): _description_

        Returns:
            dict: _description_
        """
        raw = self._mt5.symbol_info_tick(symbol)
        if raw is None:
            raise _RetryException()
        return raw

    @mt5_connection
    def bars(
        self,
        symbol,
        timeframe,
        since: int | datetime | None = 0,
        to: int | datetime | None = 1_000,
        **kwargs,
    ) -> list[list]:
        """_summary_

        Args:
            symbol (_type_): _description_
            timeframe (_type_): _description_
            since (int | datetime | None, optional): _description_. Defaults to 0.
            to (int | datetime | None, optional): _description_. Defaults to 1_000.

        Returns:
            list[list]: _description_
        """
        timeframe = TIMEFRAME_L2M[timeframe]

        if isinstance(since, int):
            raw = self._mt5.copy_rates_from_pos(symbol, timeframe, since, to)

        elif isinstance(to, int):
            raw = self._mt5.copy_rates_from(symbol, timeframe, since, to)
        else:
            raw = self._mt5.copy_rates_range(symbol, timeframe, since, to)

        if raw is None:
            raise _RetryException()
        return raw

    # Private
    # Account
    @mt5_connection
    def account(self, **kwargs) -> dict:
        """Metatrader 5 account information

        Returns:
            dict: _description_
        """
        raw = self._mt5.account_info()
        if raw is None:
            raise _RetryException()
        return raw

    # Order
    @mt5_connection
    def orders_total(
        self,
        since: datetime | None = None,
        to: datetime | None = None,
        **kwargs,
    ) -> int:
        """_summary_

        Args:
            since (datetime | None, optional): _description_. Defaults to None.
            to (datetime | None, optional): _description_. Defaults to None.

        Returns:
            int: _description_
        """
        if since is not None:
            kwargs["date_from"] = since
        if to is not None:
            kwargs["date_to"] = to

        raw = self._mt5.orders_total(**kwargs)
        if raw is None:
            raise _RetryException()

        return raw

    @mt5_connection
    def orders_get(
        self,
        id: str | None = None,
        symbol: str | None = None,
        **kwargs,
    ) -> list[dict]:
        """_summary_

        Args:
            id (str | None, optional): _description_. Defaults to None.
            symbol (str | None, optional): _description_. Defaults to None.

        Returns:
            list[dict]: _description_
        """
        if id is not None:
            kwargs["ticket"] = int(id)
        if symbol is not None:
            kwargs["symbol"] = symbol

        raws = self._mt5.orders_get(**kwargs)

        # Return None to retry
        if raws is None:
            raise _RetryException()

        return [self._order_parse_response(raw) for raw in raws]

    def order_open(self, order: "MetaTraderOrder") -> dict:
        """_summary_

        Args:
            order (MetaTraderOrder): _description_

        Raises:
            NotImplementedError: _description_

        Returns:
            dict: _description_
        """
        match order.type:
            case OrderType.Limit:
                price = order.limit_price
            case OrderType.Stop:
                price = order.stop_price
            case OrderType.Market:
                tick = self.tick_get(order.data.symbol)
                price = tick.ask if order.is_long else tick.bid
            case _:
                raise NotImplementedError(
                    f"Open order type {order.type} is not implement yet"
                )

        type = MT5.ORDER_TYPE_BUY if order.is_long else MT5.ORDER_TYPE_SELL

        return self.do_order_open(
            symbol=order.data.symbol,
            type=type,
            size=order.size,
            price=price,
            sl=order.sl,
            tp=order.tp,
            tag=order.tag,
        )

    @mt5_connection
    def do_order_open(
        self,
        symbol: str,
        size: float,
        type: int,
        price: float,
        sl: float = None,
        tp: float = None,
        tag: str | None = None,
        deviation: int = 10,
        **kwargs,
    ) -> dict:
        """_summary_

        Args:
            symbol (str): _description_
            size (float): _description_
            type (int): _description_
            price (float): _description_
            sl (float, optional): _description_. Defaults to None.
            tp (float, optional): _description_. Defaults to None.
            tag (str, optional): _description_. Defaults to "".
            deviation (int, optional): _description_. Defaults to 10.

        Returns:
            dict: _description_
        """
        request = self._parse_trade_request(
            symbol=symbol,
            size=size,
            type=type,
            price=price,
            sl=sl,
            tp=tp,
            tag=tag,
            deviation=deviation,
        )
        raw = self._mt5.order_send(request)

        # Retry
        if raw is None:
            raise _RetryException()

        raw = self._parse_trade_send_response(raw)
        if raw.code != 0:
            raise LetLiveOrderInvalidException(raw.error, raw=raw)

        return raw

    def _parse_trade_request(
        self,
        symbol: str | None = None,
        size: float | None = None,
        type: int | None = None,
        price: float | None = None,
        sl: float | None = None,
        tp: float | None = None,
        tag: str | None = None,
        position: int | None = None,
        deviation: int = 10,
        action: int = MT5.TRADE_ACTION_DEAL,
        type_time: int = MT5.ORDER_TIME_GTC,
        type_filling: int = MT5.ORDER_FILLING_IOC,
    ):
        request = {
            "magic": self._magic,
            "action": action,
            "deviation": deviation,
            "type_time": type_time,
            "type_filling": type_filling,
        }

        if symbol is not None:
            request["symbol"] = symbol
        if size is not None:
            request["volume"] = abs(size)
        if type is not None:
            request["type"] = type
        if price is not None:
            request["price"] = price
        if sl is not None:
            request["sl"] = sl
        if tp is not None:
            request["tp"] = tp
        if tag is not None:
            request["comment"] = tag
        if position is not None:
            request["position"] = position

        if __debug__:
            logger.info("New trade request: %s", request)

        return request

    def order_update(self, order: "MetaTraderOrder", sl=None, tp=None, **kwargs):
        raise NotImplementedError

    def order_close(self, order: "MetaTraderOrder", **kwargs):
        raise NotImplementedError

    def _parse_trade_send_response(self, raw) -> dict:
        if raw is None:
            raise RuntimeError("Trade response is None")

        response = Box(dict(raw._asdict()))

        # Error code
        response.code = response.retcode
        if response.code == MT5.TRADE_RETCODE_DONE:
            response.code = 0
        elif response.code == MT5.TRADE_RETCODE_NO_CHANGES:
            logger.warning(
                "Trade response nothing changes code %s %s", response.code, response
            )
            response.code = 0
        if response.code != 0:
            response.error = response.comment

        # Order
        if hasattr(response, "order"):
            response.order_id = response.order

        # Execution
        if hasattr(response, "deal"):
            response.execution_id = response.deal

        if __debug__:
            logger.info("New order response: %s", response)

        return response

    @mt5_connection
    def orders_history_total(
        self,
        since: datetime | None = None,
        to: datetime | None = None,
        **kwargs,
    ) -> int:
        """_summary_

        Args:
            since (datetime | None, optional): _description_. Defaults to None.
            to (datetime | None, optional): _description_. Defaults to None.

        Returns:
            int: _description_
        """
        if since is not None:
            kwargs["date_from"] = since
        if to is not None:
            kwargs["date_to"] = to

        raw = self._mt5.history_orders_get(**kwargs)
        if raw is None:
            raise _RetryException()
        return raw

    @mt5_connection
    def orders_history_get(
        self,
        id: str | None = None,
        position_id: str | None = None,
        since: datetime | None = None,
        to: datetime | None = None,
        **kwargs,
    ) -> list[dict]:
        """_summary_

        Args:
            id (str | None, optional): _description_. Defaults to None.
            position_id (str | None, optional): _description_. Defaults to None.
            since (datetime | None, optional): _description_. Defaults to None.
            to (datetime | None, optional): _description_. Defaults to None.

        Returns:
            list[dict]: _description_
        """
        if id is not None:
            kwargs["ticket"] = int(id)
        if position_id is not None:
            kwargs["position"] = int(position_id)
        if since is not None:
            kwargs["date_from"] = since
        if to is not None:
            kwargs["date_to"] = to

        raws = self._mt5.history_orders_get(**kwargs)

        # Retry
        if raws is None:
            raise _RetryException()

        return [self._order_parse_response(raw) for raw in raws]

    def _order_parse_response(self, raw):
        response = Box(dict(raw._asdict()))
        return response

    # Execution
    @mt5_connection
    def executions_total(
        self,
        since: datetime | None = None,
        to: datetime | None = None,
        **kwargs,
    ) -> int:
        """_summary_

        Args:
            since (datetime | None, optional): _description_. Defaults to None.
            to (datetime | None, optional): _description_. Defaults to None.

        Returns:
            int: _description_
        """
        if since is not None:
            kwargs["date_from"] = since
        if to is not None:
            kwargs["date_to"] = to

        raw = self._mt5.history_deals_total(**kwargs)
        if raw is None:
            raise _RetryException()
        return raw

    @mt5_connection
    def executions_get(
        self,
        position_id: str | None = None,
        search: str | None = None,
        **kwargs,
    ) -> list[dict]:
        """_summary_

        Args:
            position_id (str | None, optional): _description_. Defaults to None.
            search (str | None, optional): _description_. Defaults to None.

        Returns:
            list[dict]: _description_
        """
        if position_id is not None:
            kwargs["position"] = int(position_id)
        if search is not None:
            kwargs["group"] = search

        raws = self._mt5.history_deals_get(**kwargs)

        # Retry
        if raws is None:
            raise _RetryException()

        # May be wrong account when position exist but no execution
        if not raws and position_id is not None:
            logger.warning(
                "Execution retry check connection when position=%s exist but no execution",
                position_id,
            )

            # Retry check mt5 connection
            raise _RetryException()

        if __debug__:
            logger.debug("Raw executions: %s", raws)

        return [self._execution_parse_response(raw) for raw in raws]

    @mt5_connection
    def execution_get(self, id: str, **kwargs) -> dict:
        """_summary_

        Args:
            id (str): _description_

        Returns:
            dict: _description_
        """
        if id is not None:
            kwargs["ticket"] = int(id)

        raws = self._mt5.history_deals_get(**kwargs)

        # Retry
        if raws is None:
            raise _RetryException()

        if __debug__:
            logger.debug("Raw execution: %s", raws)

        return self._execution_parse_response(raws[0])

    def _execution_parse_response(self, raw):
        # Store
        self._executions_stored[raw.ticket] = raw

        raw = Box(dict(raw._asdict()))
        return raw

    # Position
    @mt5_connection
    def positions_total(
        self,
        since: datetime | None = None,
        to: datetime | None = None,
        **kwargs,
    ) -> int:
        """_summary_

        Args:
            since (datetime | None, optional): _description_. Defaults to None.
            to (datetime | None, optional): _description_. Defaults to None.

        Returns:
            int: _description_
        """
        if since is not None:
            kwargs["date_from"] = since
        if to is not None:
            kwargs["date_to"] = to

        raw = self._mt5.positions_total(**kwargs)
        if raw is None:
            raise _RetryException()
        return raw

    @mt5_connection
    def positions_get(
        self,
        id: str = None,
        symbol: str = None,
        **kwargs,
    ) -> list[dict]:
        """_summary_

        Args:
            id (str, optional): _description_. Defaults to None.
            symbol (str, optional): _description_. Defaults to None.

        Returns:
            list[dict]: _description_
        """
        if id is not None:
            kwargs.update(ticket=int(id))
        if symbol is not None:
            kwargs.update(symbol=symbol)

        raws = self._mt5.positions_get(**kwargs)

        # Retry
        if raws is None:
            raise _RetryException()

        return [self._position_parse_response(raw) for raw in raws]

    def position_update(
        self,
        position: "MetaTraderPosition",
        sl: float | None = None,
        tp: float | None = None,
        **kwargs,
    ) -> dict:
        """_summary_

        Args:
            position (MetaTraderPosition): _description_
            sl (float | None, optional): _description_. Defaults to None.
            tp (float | None, optional): _description_. Defaults to None.

        Returns:
            dict: _description_
        """
        return self.do_position_update(
            id=int(position.id),
            symbol=position.data.symbol,
            sl=sl,
            tp=tp,
            **kwargs,
        )

    @mt5_connection
    def do_position_update(
        self,
        id: int,
        # symbol: str,
        sl: float | None = None,
        tp: float | None = None,
        **kwargs,
    ) -> dict:
        """_summary_

        Args:
            id (int): _description_
            sl (float | None, optional): _description_. Defaults to None.
            tp (float | None, optional): _description_. Defaults to None.

        Returns:
            dict: _description_
        """
        request = self._parse_trade_request(
            position=id,
            # symbol=symbol,
            sl=sl,
            tp=tp,
            action=MT5.TRADE_ACTION_SLTP,
            **kwargs,
        )
        raw = self._mt5.order_send(request)

        # Retry
        if raw is None:
            raise _RetryException()

        return self._parse_trade_send_response(raw)

    def position_close(self, position: "MetaTraderPosition", **kwargs) -> dict:
        """Close a position

        Args:
            position (MetaTraderPosition): _description_

        Returns:
            dict: _description_
        """
        tick = self.tick_get(position.data.symbol)
        price = tick.ask if position.is_long else tick.bid

        # Opposite with position side
        type = MT5.ORDER_TYPE_SELL if position.is_long else MT5.ORDER_TYPE_BUY

        return self.do_position_close(
            id=int(position.id),
            symbol=position.data.symbol,
            type=type,
            size=position.size,
            price=price,
            **kwargs,
        )

    @mt5_connection
    def do_position_close(
        self,
        id: int,
        symbol: str,
        type: int,
        price: float,
        size: float,
        **kwargs,
    ) -> dict:
        request = self._parse_trade_request(
            position=id,
            symbol=symbol,
            type=type,
            price=price,
            size=size,
            **kwargs,
        )
        raw = self._mt5.order_send(request)

        # Retry
        if raw is None:
            raise _RetryException()

        return self._parse_trade_send_response(raw)

    def _position_parse_response(self, raw) -> dict:
        return Box(dict(raw._asdict()))

    # Transaction
    def _load_history_transactions(self):
        to = datetime.now()  # + timedelta(days=1)

        # History orders
        raws = self._mt5.history_orders_get(self._load_history_since, to)
        if raws is None:
            raise _RetryException()

        self._exchange.on_orders_event(old=raws)

        # # Positions
        # positions = self._mt5.positions_get(self._load_history_since, to)
        # if positions:
        #     self._exchange.on_positions_new(positions)

        # TODO: generate history order from
        # # History positions
        # positions = self._mt5.history_()
        # self._exchange.on_positions_old(positions)

    def _check_transaction_events(self):
        if not self._exchange:
            return

        # Deals
        deals = self._check_deals()
        if deals:
            self._exchange.on_executions_event(deals)

        # Orders
        orders_new, orders_old = self._check_orders()
        if orders_new or orders_old:
            self._exchange.on_orders_event(new=orders_new, old=orders_old)

        # Positions
        positions_new, positions_old = self._check_positions()
        if positions_new or positions_old:
            self._exchange.on_positions_event(new=positions_new, old=positions_old)

    # Deal
    @mt5_connection
    def _check_deals(self, **kwargs) -> list | bool | None:
        to = datetime.now()  # + timedelta(days=1)

        deal_total = self._mt5.history_deals_total(self._deal_time_checked, to)

        # Retry
        if deal_total is None:
            raise _RetryException()

        # No thing new in deal
        if deal_total <= 0:
            return False

        raws = self._mt5.history_deals_get(self._deal_time_checked, to)

        # Retry
        if raws is None:
            raise _RetryException()

        if len(raws) > 0:
            # Update last check time +1 second
            self._deal_time_checked = datetime.fromtimestamp(raws[-1].time + 1)

            # Store
            for raw in raws:
                self._executions_stored[raw.ticket] = raw

        return raws

    # Order
    @mt5_connection
    def _check_orders(self, **kwargs) -> tuple | None:
        order_total = self._mt5.orders_total()

        # Retry
        if order_total is None:
            raise _RetryException()

        # No thing new in order
        if order_total <= 0 and len(self._orders_stored) == 0:
            return None, None

        raws = self._mt5.orders_get()

        # Retry
        if raws is None:
            raise _RetryException()

        tickets = [raw.ticket for raw in raws]

        removed_orders = [
            raw for raw in self._orders_stored.values() if raw.ticket not in tickets
        ]

        added_orders = []
        for raw in raws:
            if raw.ticket in self._orders_stored:
                stored = self._orders_stored[raw.ticket]
                if (
                    raw.sl == stored.sl
                    and raw.tp == stored.tp
                    and raw.volume_current == stored.volume_current
                    and raw.price_open == stored.price_open
                    and raw.price_stoplimit == stored.price_stoplimit
                ):
                    continue

            added_orders.append(raw)
        self._orders_stored = {raw.ticket: raw for raw in raws}
        return added_orders, removed_orders

    # Trade
    @mt5_connection
    def _check_positions(self, **kwargs) -> tuple | None:
        positions_total = self._mt5.positions_total()

        # Retry
        if positions_total is None:
            raise _RetryException()

        # No thing new in trade
        if positions_total <= 0 and not bool(self._positions_stored):
            return None, None

        raws = self._mt5.positions_get()

        # Retry
        if raws is None:
            raise _RetryException()

        # May be wrong account when position exist but no execution
        if not raws and bool(self._positions_stored):
            position_exit = list(self._positions_stored.values())[-1]
            executions = [
                e
                for e in self._executions_stored.values()
                if e.position_id == position_exit.ticket
            ]

            # Execution is not close of position
            if not executions or executions[-1].type == position_exit.type:
                logger.warning(
                    "Positions retry check connection when position=%s exited but no execution",
                    position_exit.ticket,
                )
            raise _RetryException()

        tickets = [raw.ticket for raw in raws]
        raws = {raw.ticket: self._position_parse_response(raw) for raw in raws}

        removed_positions = [
            raw for raw in self._positions_stored.values() if raw.ticket not in tickets
        ]

        added_positions = []
        for raw in raws.values():
            if raw.ticket in self._positions_stored:
                stored = self._positions_stored[raw.ticket]
                if (
                    raw.time_update == stored.time_update
                    and raw.sl == stored.sl
                    and raw.tp == stored.tp
                    and raw.volume == stored.volume
                    # and raw.price_open == stored.price_open
                ):
                    continue

            added_positions.append(raw)

        self._positions_stored = raws
        return added_positions, removed_positions

    # Bypass pickle
    def __copy__(self):
        return self.__class__._singleton

    def __deepcopy__(self, memo):
        return self.__class__._singleton



import logging

from lettrade.account import Account

from .api import LiveAPI
from .trade import LivePosition

logger = logging.getLogger(__name__)


class LiveAccount(Account):
    _api: LiveAPI
    _account: object

    _balance: float
    _equity: float

    def __init__(self, api: LiveAPI, **kwargs) -> None:
        """Account for live trading

        Args:
            api (LiveAPI): _description_
            **kwargs (dict, optional): Mirror of [lettrade.account.Account()](site:/reference/account/account/#lettrade.account.account.Account).
        """
        super().__init__(**kwargs)
        self._api = api

        self._balance = 0.0
        self._equity = 0.0

    # def __repr__(self):
    #     return "<LiveAccount " + str(self) + ">"

    def start(self):
        """Start live account by load account info from API"""
        self.account_refresh()
        self._margin = self._account.margin
        self._leverage = self._account.leverage

    # def next(self):
    #     """Live account next"""
    #     self.account_refresh()
    #     return super().next()

    def pl(self, size, entry_price: float, exit_price: float | None = None) -> float:
        if exit_price is None:
            exit_price = self._exchange.data.l.open[0]

        pl = size * (exit_price - entry_price)

        return pl * 100_000

    def account_refresh(self):
        """Refresh account balance"""
        self._account = self._api.account()

        if __debug__:
            logger.debug("Account: %s", str(self._account))

        self._balance = self._account.balance
        self._equity = self._account.equity

    def on_positions(self, positions: list[LivePosition]):
        self.account_refresh()
        return super().on_positions(positions)

    @property
    def equity(self) -> float:
        return self._equity



import pandas as pd

from lettrade.plot.helper import plot_merge


def plot_ichimoku(
    dataframe: pd.DataFrame,
    tenkan_sen="tenkan_sen",
    kijun_sen="kijun_sen",
    senkou_span_a="senkou_span_a",
    senkou_span_b="senkou_span_b",
    chikou_span="chikou_span",
    width=1,
    tenkan_sen_color="#33BDFF",
    kijun_sen_color="#D105F5",
    senkou_span_a_color="#228B22",
    senkou_span_b_color="#FF3342",
    chikou_span_color="#F1F316",
    filter: pd.Series | None = None,
) -> dict:
    """_summary_

    Args:
        dataframe (pd.DataFrame): _description_
        tenkan_sen (str, optional): _description_. Defaults to "tenkan_sen".
        kijun_sen (str, optional): _description_. Defaults to "kijun_sen".
        senkou_span_a (str, optional): _description_. Defaults to "senkou_span_a".
        senkou_span_b (str, optional): _description_. Defaults to "senkou_span_b".
        chikou_span (str, optional): _description_. Defaults to "chikou_span".
        width (int, optional): _description_. Defaults to 1.
        tenkan_sen_color (str, optional): _description_. Defaults to "#33BDFF".
        kijun_sen_color (str, optional): _description_. Defaults to "#D105F5".
        senkou_span_a_color (str, optional): _description_. Defaults to "#228B22".
        senkou_span_b_color (str, optional): _description_. Defaults to "#FF3342".
        chikou_span_color (str, optional): _description_. Defaults to "#F1F316".
        filter (pd.Series | None, optional): _description_. Defaults to None.

    Returns:
        dict: _description_
    """
    if filter is not None:
        df_name = dataframe.name
        dataframe = dataframe.loc[filter]
        object.__setattr__(dataframe, "name", df_name)

    items = []

    if tenkan_sen is not None:
        items.append(
            dict(
                type="scatter",
                x=dataframe.index,
                y=dataframe[tenkan_sen],
                name=tenkan_sen,
                mode="lines",
                line=dict(color=tenkan_sen_color, width=width),
            )
        )
    if kijun_sen is not None:
        items.append(
            dict(
                type="scatter",
                x=dataframe.index,
                y=dataframe[kijun_sen],
                name=kijun_sen,
                mode="lines",
                line=dict(color=kijun_sen_color, width=width),
            )
        )
    if senkou_span_a is not None:
        items.append(
            dict(
                type="scatter",
                x=dataframe.index,
                y=dataframe[senkou_span_a],
                name=senkou_span_a,
                mode="lines",
                line=dict(color=senkou_span_a_color, width=width),
            )
        )
    if senkou_span_b is not None:
        items.append(
            dict(
                type="scatter",
                x=dataframe.index,
                y=dataframe[senkou_span_b],
                name=senkou_span_b,
                mode="lines",
                fill="tonexty",
                line=dict(color=senkou_span_b_color, width=width),
            )
        )
    if chikou_span is not None:
        items.append(
            dict(
                type="scatter",
                x=dataframe.index,
                y=dataframe[chikou_span],
                name=chikou_span,
                mode="lines",
                line=dict(color=chikou_span_color, width=width),
            )
        )

    return {f"{dataframe.name}": dict(items=items)}


def plot_bollinger_bands(
    dataframe: pd.DataFrame,
    upper: str | None = "upper",
    basis: str | None = "basis",
    lower: str | None = "lower",
    width: int = 1,
    upper_color: str = "#33BDFF",
    basis_color: str = "#D105F5",
    lower_color: str = "#33BDFF",
    filter: pd.Series | None = None,
) -> dict:
    """_summary_

    Args:
        dataframe (pd.DataFrame): _description_
        upper (str, optional): Column name. `None` mean skip plot. Defaults to "upper".
        basis (str, optional): Column name. `None` mean skip plot. Defaults to "basis".
        lower (str, optional): Column name. `None` mean skip plot. Defaults to "lower".
        width (int, optional): _description_. Defaults to 1.
        upper_color (str, optional): Color code. Defaults to "#33BDFF".
        lower_color (str, optional): Color code. Defaults to "#33BDFF".
        basis_color (str, optional): Color code. Defaults to "#D105F5".
        filter (pd.Series | None, optional): _description_. Defaults to None.

    Returns:
        dict: _description_
    """
    if filter is not None:
        df_name = dataframe.name
        dataframe = dataframe.loc[filter]
        object.__setattr__(dataframe, "name", df_name)

    items = []

    if upper is not None:
        items.append(
            dict(
                type="scatter",
                x=dataframe.index,
                y=dataframe[upper],
                name=upper,
                mode="lines",
                line=dict(color=upper_color, width=width),
            )
        )
    if basis is not None:
        items.append(
            dict(
                type="scatter",
                x=dataframe.index,
                y=dataframe[basis],
                name=basis,
                mode="lines",
                line=dict(color=basis_color, width=width),
            )
        )
    if lower is not None:
        items.append(
            dict(
                type="scatter",
                x=dataframe.index,
                y=dataframe[lower],
                name=lower,
                mode="lines",
                line=dict(color=lower_color, width=width),
            )
        )
    return {f"{dataframe.name}": dict(items=items)}


plot_keltner_channel = plot_bollinger_bands


def plot_parabolic_sar(
    dataframe: pd.DataFrame,
    long: str | None = "long",
    short: str | None = "short",
    width: int = 1,
    long_color: str = "#33BDFF",
    short_color: str = "#D105F5",
    filter: pd.Series | None = None,
) -> dict:
    """_summary_

    Args:
        dataframe (pd.DataFrame): _description_
        long (str | None, optional): _description_. Defaults to "long".
        short (str | None, optional): _description_. Defaults to "short".
        width (int, optional): _description_. Defaults to 1.
        long_color (str, optional): _description_. Defaults to "#33BDFF".
        short_color (str, optional): _description_. Defaults to "#D105F5".
        filter (pd.Series | None, optional): _description_. Defaults to None.

    Returns:
        dict: _description_
    """
    if filter is not None:
        df_name = dataframe.name
        dataframe = dataframe.loc[filter]
        object.__setattr__(dataframe, "name", df_name)

    config = dict()

    if long is not None:
        plot_merge(
            config,
            plot_mark(
                series=long,
                color=long_color,
                name="psar_long",
                width=width,
                dataframe=dataframe,
            ),
        )
    if short is not None:
        plot_merge(
            config,
            plot_mark(
                series=short,
                color=short_color,
                name="psar_short",
                width=width,
                dataframe=dataframe,
            ),
        )
    return config


def plot_line(
    series: pd.Series | str,
    color: str = "#ffee58",
    width: int = 1,
    name: str | None = None,
    mode: str = "lines",
    fullfill: bool = False,
    dataframe: pd.DataFrame | None = None,
    filter: pd.Series | None = None,
    **kwargs,
) -> dict:
    """_summary_

    Args:
        series (pd.Series | str): _description_
        color (str, optional): _description_. Defaults to "#ffee58".
        width (int, optional): _description_. Defaults to 1.
        name (str | None, optional): _description_. Defaults to None.
        mode (str, optional): _description_. Defaults to "lines".
        fullfill (bool, optional): _description_. Defaults to False.
        dataframe (pd.DataFrame | None, optional): _description_. Defaults to None.
        filter (pd.Series | None, optional): _description_. Defaults to None.

    Returns:
        dict: _description_
    """
    if isinstance(series, str):
        series = dataframe[series]

    if filter is not None:
        series = series.loc[filter]

    config = dict(
        items=[
            dict(
                type="scatter",
                x=series.index,
                y=series,
                line=dict(color=color, width=width),
                name=name or series.name,
                mode=mode,
                fullfill=fullfill,
                **kwargs,
            )
        ]
    )
    if dataframe is None:
        return config

    return {f"{dataframe.name}": config}


def plot_lines(
    *serieses: list[pd.Series | str],
    color: str = "#ffee58",
    width: int = 1,
    name: str | None = None,
    mode: str = "lines",
    fullfill: bool = False,
    dataframe: pd.DataFrame | None = None,
    **kwargs,
) -> dict:
    """_summary_

    Args:
        color (str, optional): _description_. Defaults to "#ffee58".
        width (int, optional): _description_. Defaults to 1.
        name (str | None, optional): _description_. Defaults to None.
        mode (str, optional): _description_. Defaults to "lines".
        fullfill (bool, optional): _description_. Defaults to False.
        dataframe (pd.DataFrame | None, optional): _description_. Defaults to None.

    Returns:
        dict: _description_
    """
    result = {}
    for series in serieses:
        plot_merge(
            result,
            plot_line(
                series=series,
                color=color,
                width=width,
                name=name,
                mode=mode,
                fullfill=fullfill,
                dataframe=dataframe,
                **kwargs,
            ),
        )
    return result


def plot_mark(
    series: pd.Series | str,
    color: str = "#ffee58",
    width: int = 1,
    mode: str = "markers",
    name: str | None = None,
    fullfill: bool = False,
    dataframe: pd.DataFrame | None = None,
    filter: pd.Series | None = None,
    **kwargs,
) -> dict:
    """_summary_

    Args:
        series (pd.Series | str): _description_
        color (str, optional): _description_. Defaults to "#ffee58".
        width (int, optional): _description_. Defaults to 1.
        mode (str, optional): _description_. Defaults to "markers".
        name (str | None, optional): _description_. Defaults to None.
        fullfill (bool, optional): _description_. Defaults to False.
        dataframe (pd.DataFrame | None, optional): _description_. Defaults to None.
        filter (pd.Series | None, optional): _description_. Defaults to None.

    Returns:
        dict: _description_
    """

    return plot_line(
        series=series,
        color=color,
        width=width,
        mode=mode,
        name=name,
        fullfill=fullfill,
        dataframe=dataframe,
        filter=filter,
        **kwargs,
    )


def plot_candlestick(
    dataframe: pd.DataFrame,
    name: str = "Candlestick",
    width: int = 1,
    increasing_line_color="#26c6da",
    decreasing_line_color="#ab47bc",
    row: int = 1,
    col: int = 1,
    filter: pd.Series | None = None,
    **kwargs,
) -> dict:
    """_summary_

    Args:
        dataframe (pd.DataFrame): _description_
        name (str, optional): _description_. Defaults to "Candlestick".
        width (int, optional): _description_. Defaults to 1.
        increasing_line_color (str, optional): _description_. Defaults to "#26c6da".
        decreasing_line_color (str, optional): _description_. Defaults to "#ab47bc".
        row (int, optional): _description_. Defaults to 1.
        col (int, optional): _description_. Defaults to 1.
        filter (pd.Series | None, optional): _description_. Defaults to None.

    Returns:
        dict: _description_
    """
    if filter is not None:
        df_name = dataframe.name
        dataframe = dataframe.loc[filter]
        object.__setattr__(dataframe, "name", df_name)

    config = dict(
        items=[
            dict(
                type="candlestick",
                x=dataframe.index,
                open=dataframe["open"],
                high=dataframe["high"],
                low=dataframe["low"],
                close=dataframe["close"],
                name=name,
                line=dict(width=width),
                increasing_line_color=increasing_line_color,
                decreasing_line_color=decreasing_line_color,
                hoverinfo="text",
                hovertext=name,
                row=row,
                col=col,
                **kwargs,
            ),
        ]
    )

    return {f"{dataframe.name}": config}



import logging

from lettrade import BotStatistic, Commander, Plotter, Strategy
from lettrade.exchange.live import (
    LetTradeLive,
    LetTradeLiveBot,
    LiveAccount,
    LiveDataFeed,
    LiveDataFeeder,
    LiveExchange,
    let_live,
)

from .api import MetaTraderAPI
from .trade import MetaTraderExecution, MetaTraderOrder, MetaTraderPosition

logger = logging.getLogger(__name__)


class MetaTraderDataFeed(LiveDataFeed):
    """DataFeed for MetaTrader"""

    _api_cls: type[MetaTraderAPI] = MetaTraderAPI
    _bar_datetime_unit: str = "s"

    # def push(self, rows: list):
    #     if isinstance(rows, np.ndarray):
    #         df = pd.DataFrame(
    #             rows,
    #             columns=[
    #                 "time",
    #                 "open",
    #                 "high",
    #                 "low",
    #                 "close",
    #                 "tick_volume",
    #             ],
    #         )
    #         df.rename(
    #             columns={
    #                 "time": "datetime",
    #                 "tick_volume": "volume",
    #             },
    #             inplace=True,
    #         )
    #         df["datetime"] = pd.to_datetime(df["datetime"], unit="s")
    #         df.set_index("datetime", inplace=True)
    #         print(df, df.dtypes)

    #     return super().push(df)


class MetaTraderDataFeeder(LiveDataFeeder):
    """DataFeeder for MetaTrader"""

    _api_cls: type[MetaTraderAPI] = MetaTraderAPI
    _data_cls: type[MetaTraderDataFeed] = MetaTraderDataFeed


class MetaTraderAccount(LiveAccount):
    """Account for MetaTrader"""


class MetaTraderExchange(LiveExchange):
    """MetaTrade 5 exchange module for `lettrade`"""

    _execution_cls: type[MetaTraderExecution] = MetaTraderExecution
    _order_cls: type[MetaTraderOrder] = MetaTraderOrder
    _position_cls: type[MetaTraderPosition] = MetaTraderPosition

    def on_order(
        self,
        order: MetaTraderOrder,
        broadcast: bool | None = True,
        **kwargs,
    ) -> None:
        if not order.is_real:
            if order.id in self.history_orders:
                del self.history_orders[order.id]

        return super().on_order(order=order, broadcast=broadcast, **kwargs)


class LetTradeMetaTraderBot(LetTradeLiveBot):
    """LetTradeBot for MetaTrader"""


class LetTradeMetaTrader(LetTradeLive):
    """Help to maintain MetaTrader bots"""

    _data_cls: type[MetaTraderDataFeed] = MetaTraderDataFeed

    def __init__(
        self,
        feeder: type[MetaTraderDataFeeder] = MetaTraderDataFeeder,
        exchange: type[MetaTraderExchange] = MetaTraderExchange,
        account: type[MetaTraderAccount] = MetaTraderAccount,
        **kwargs,
    ) -> None:
        """_summary_

        Args:
            feeder (Type[MetaTraderDataFeeder], optional): _description_. Defaults to MetaTraderDataFeeder.
            exchange (Type[MetaTraderExchange], optional): _description_. Defaults to MetaTraderExchange.
            account (Type[MetaTraderAccount], optional): _description_. Defaults to MetaTraderAccount.
        """
        super().__init__(
            feeder=feeder,
            exchange=exchange,
            account=account,
            **kwargs,
        )


def let_metatrader(
    datas: set[set[str]],
    strategy: type[Strategy],
    *,
    mt5_login: int,
    mt5_password: str,
    mt5_server: str,
    mt5_wine: str | None = None,
    mt5_path: str | None = None,
    feeder: type[MetaTraderDataFeeder] = MetaTraderDataFeeder,
    exchange: type[MetaTraderExchange] = MetaTraderExchange,
    account: type[MetaTraderAccount] = MetaTraderAccount,
    commander: type[Commander] | None = None,
    stats: type[BotStatistic] = BotStatistic,
    plotter: type[Plotter] | None = None,
    bot: type[LetTradeMetaTraderBot] = LetTradeMetaTraderBot,
    lettrade: type[LetTradeMetaTrader] = LetTradeMetaTrader,
    api: type[MetaTraderAPI] = MetaTraderAPI,
    **kwargs,
) -> LetTradeMetaTrader:
    """Help to build `LetTradeMetaTrader`

    Args:
        datas (set[set[str]]): _description_
        strategy (type[Strategy]): _description_
        mt5_login (int): _description_
        mt5_password (str): _description_
        mt5_server (str): _description_
        mt5_wine (str | None, optional): WineHQ execute path. Defaults to None.
        feeder (type[MetaTraderDataFeeder], optional): _description_. Defaults to MetaTraderDataFeeder.
        exchange (type[MetaTraderExchange], optional): _description_. Defaults to MetaTraderExchange.
        account (type[MetaTraderAccount], optional): _description_. Defaults to MetaTraderAccount.
        commander (type[Commander] | None, optional): _description_. Defaults to None.
        stats (type[BotStatistic], optional): _description_. Defaults to BotStatistic.
        plotter (type[Plotter] | None, optional): _description_. Defaults to None.
        bot (type[LetTradeMetaTraderBot], optional): _description_. Defaults to LetTradeMetaTraderBot.
        lettrade (type[LetTradeMetaTrader], optional): _description_. Defaults to LetTradeMetaTrader.
        api (type[MetaTraderAPI], optional): _description_. Defaults to MetaTraderAPI.

    Returns:
        LetTradeMetaTrader: _description_
    """
    api_kwargs: dict = kwargs.setdefault("api_kwargs", {})
    api_kwargs.update(
        login=int(mt5_login),
        password=mt5_password,
        server=mt5_server,
        wine=mt5_wine,
        path=mt5_path,
    )

    return let_live(
        strategy=strategy,
        datas=datas,
        feeder=feeder,
        exchange=exchange,
        account=account,
        commander=commander,
        stats=stats,
        plotter=plotter,
        bot=bot,
        lettrade=lettrade,
        api=api,
        **kwargs,
    )



"""Exchange implements

# Stable Exchanges
* [BackTest](./backtest/index.md): Backtesting exchange
* [MetaTrader](./metatrader/index.md): MetaTrader 5 live data and trading
* [CCXT](./ccxt/index.md): CCXT CryptoCurrency live data and trading
"""

from .base import OrderState, OrderType, PositionState, TradeSide
from .error import *
from .exchange import Exchange
from .execution import Execution
from .order import Order, OrderResult, OrderResultError, OrderResultOk
from .position import Position, PositionResult, PositionResultError, PositionResultOk



import functools
import logging
import time
from datetime import datetime
from multiprocessing.managers import BaseManager
from typing import TYPE_CHECKING, Literal

import ccxt
from box import Box
from ccxt.base.errors import RequestTimeout

from lettrade.exchange.live import LetLiveOrderInvalidException, LiveAPI

if TYPE_CHECKING:
    from .ccxt import CCXTExchange
    from .trade import CCXTOrder, CCXTPosition


logger = logging.getLogger(__name__)


def ccxt_connection(api_function):
    @functools.wraps(api_function)
    def wrapper(self: "CCXTAPIExchange", *args, api_retry: int = 3, **kwargs):
        while api_retry > 0:
            try:
                return api_function(self, *args, **kwargs)
            except RequestTimeout:
                pass

            logger.warning("Retry functon %s", api_function)

            time.sleep(1)
            api_retry -= 1

    return wrapper


class CCXTAPIExchange:
    """Single instance across multiprocessing. Help pickle-able result and send across multiprocessing"""

    _exchange: ccxt.Exchange

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "_singleton"):
            cls._singleton = object.__new__(cls)
        return cls._singleton

    def __init__(
        self,
        exchange: str,
        key: str,
        secret: str,
        options: dict | None = None,
        type: Literal["spot", "margin", "future"] = "spot",
        sandbox: bool = True,
        verbose: bool = False,
        **kwargs,
    ) -> None:
        config = dict(
            apiKey=key,
            secret=secret,
            enableRateLimit=True,
            defaultType=type,
            options={
                "sandboxMode": sandbox,
                "warnOnFetchOpenOrdersWithoutSymbol": False,
                "tradesLimit": 1,
                "ordersLimit": 1,
                "OHLCVLimit": 1,
            },
        )
        config.update(kwargs)

        if options is not None:
            config["options"].update(options)

        self._exchange = getattr(ccxt, exchange)(config)

        # Must call sanbox function instead of option sandboxMode
        self._exchange.set_sandbox_mode(sandbox)
        self._exchange.verbose = verbose
        logger.info("Starting exchange class: %s", self._exchange)

    def start(self):
        self._exchange.load_markets()

    def stop(self):
        """"""
        # self._exchange.close()

    # def __getattr__(self, attr: str) -> ccxt.Exchange:
    #     # if hasattr(self._exchange, attr):
    #     return getattr(self._exchange, attr)

    # Public functions
    def fetch_trades(self, symbol, since=None, limit=None, params={}):
        return self._exchange.watch_trades(symbol, since, limit, params)

    def fetch_time(self):
        return self._exchange.milliseconds()

    # Ticker
    def fetch_ticker(self, symbol, **kwargs):
        return self._exchange.fetch_ticker(symbol, **kwargs)

    def watch_ticker(self, symbol, **kwargs):
        return self._exchange.watch_ticker(symbol, **kwargs)

    def fetch_tickers(self, symbols, **kwargs):
        return self._exchange.fetch_tickers(symbols, **kwargs)

    def watch_tickers(self, symbols, **kwargs):
        return self._exchange.watch_tickers(symbols)

    # Market
    def fetch_markets(self):
        return self._exchange.markets

    # Bar
    @ccxt_connection
    def fetch_ohlcv(self, *args, **kwargs):
        return self._exchange.fetch_ohlcv(*args, **kwargs)

    @ccxt_connection
    def watch_ohlcv(self, *args, **kwargs):
        return self._exchange.watch_ohlcv(*args, **kwargs)

    # Account functions
    # balance
    def fetch_my_balance(self, params={}):
        return self._exchange.fetch_balance(params)

    def watch_my_balance(self, params={}, **kwargs):
        return self._exchange.watch_balance(params)

    # Order
    def create_my_order(self, symbol, type, side, amount, price, **kwargs):
        return self._exchange.create_order(
            symbol=symbol,
            type=type,
            side=side,
            amount=amount,
            price=price,
            **kwargs,
        )

    def fetch_my_order(self, oid, symbol):
        return self._exchange.fetch_order(oid, symbol)

    def fetch_my_orders(self, symbol=None, since=None, limit=None, params={}):
        return self._exchange.fetch_orders(symbol, since, limit, params)

    def fetch_my_open_orders(self, symbol=None, since=None, limit=None, params={}):
        return self._exchange.fetch_open_orders(symbol, since, limit, params)

    def cancel_my_order(self, id, symbol):
        return self._exchange.cancel_order(id, symbol)

    def cancel_my_orders(self, symbol=None, params={}):
        return self._exchange.cancel_all_orders(symbol, params)

    # Trade
    def fetch_my_trades(self, symbol=None, since=None, limit=None, params={}):
        return self._exchange.fetch_my_trades(symbol, since, limit, params)

    def watch_my_trades(self, symbol=None, since=None, limit=None, params={}):
        return self._exchange.watch_my_trades(symbol, since, limit, params)

    # Position
    def fetch_my_position(self, symbol: str, params={}):
        return self._exchange.fetch_position(symbol, params)

    def fetch_my_positions(self, symbols=None, params={}):
        return self._exchange.fetch_positions(symbols, params)


class CCXTAPI(LiveAPI):
    """CCXT API"""

    _ccxt: CCXTAPIExchange
    _exchange: "CCXTExchange"
    _currency: str

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "_singleton"):
            cls._singleton = object.__new__(cls)
        return cls._singleton

    @classmethod
    def multiprocess(cls, kwargs, **other_kwargs):
        BaseManager.register("CCXTAPIExchange", CCXTAPIExchange)
        manager = BaseManager()
        manager.start()
        kwargs["ccxt"] = manager.CCXTAPIExchange(**kwargs)

    def __init__(
        self,
        exchange: int,
        key: str,
        secret: str,
        currency: str = "USDT",
        ccxt: CCXTAPIExchange | None = None,
        **kwargs,
    ):
        """_summary_

        Args:
            exchange (int): _description_
            key (str): _description_
            secret (str): _description_
            ccxt (CCXTAPIExchange | None, optional): _description_. Defaults to None.
        """
        if ccxt is None:
            ccxt = CCXTAPIExchange(exchange=exchange, key=key, secret=secret, **kwargs)
        self._ccxt = ccxt
        self._currency = ccxt

    # Bypass pickle
    def __copy__(self):
        return self.__class__._singleton

    def __deepcopy__(self, memo):
        return self.__class__._singleton

    def start(self, exchange: "CCXTExchange | None" = None):
        self._exchange = exchange
        self._currency = exchange._account._currency

        self._ccxt.start()

    def stop(self):
        """"""
        # self._ccxt.stop()

    def next(self):
        pass

    def heartbeat(self):
        return True

    def market(self, symbol: str) -> dict:
        pass

    def markets(self, symbols: list[str]) -> dict:
        pass

    def tick_get(self, symbol: str) -> dict:
        pass

    # Bars
    def bars(
        self,
        symbol,
        timeframe,
        since: int | datetime | None = 0,
        to: int | datetime | None = 1_000,
        **kwargs,
    ) -> list[list]:
        return self._ccxt.fetch_ohlcv(symbol, timeframe, limit=to, **kwargs)

    ### Private
    # Account
    def account(self) -> dict:
        """"""
        raw = self._ccxt.fetch_my_balance()
        currency = raw[self._currency]
        return Box(
            balance=currency["free"],
            equity=currency["total"],
            margin=1,
            leverage=1,
            raw=raw,
        )

    #  Order
    def order_open(self, order: "CCXTOrder", **kwargs):
        """"""
        try:
            result = self._ccxt.create_my_order(
                symbol=order.data.symbol,
                type=order.type.lower(),
                side=order.side.lower(),
                amount=abs(order.size),
                price=order.place_price,
                **kwargs,
            )

            print("order_open", order, result)
            return result
        except ccxt.InvalidOrder as e:
            raise LetLiveOrderInvalidException(e.args[0]) from e

    def order_update(self, order: "CCXTOrder"):
        pass

    def order_close(self, order: "CCXTOrder"):
        """"""

    def orders_total(self):
        """"""

    def orders_get(self, **kwargs):
        """"""

    # Execution
    def executions_total(
        self,
        since: datetime | None = None,
        to: datetime | None = None,
        **kwargs,
    ) -> int:
        """"""

    def executions_get(
        self,
        position_id: str | None = None,
        search: str | None = None,
        **kwargs,
    ) -> list[dict]:
        """"""

    def execution_get(self, id: str, **kwargs) -> dict:
        """"""

    # Position
    def positions_total(
        self,
        since: datetime | None = None,
        to: datetime | None = None,
        **kwargs,
    ) -> int:
        """"""

    def positions_get(self, id: str = None, symbol: str = None, **kwargs) -> list[dict]:
        """"""

    def position_update(
        self,
        position: "CCXTPosition",
        sl: float | None = None,
        tp: float | None = None,
        **kwargs,
    ) -> dict:
        """"""

    def position_close(self, position: "CCXTPosition", **kwargs) -> dict:
        """"""



import unittest

import pandas as pd
from pandas import testing as pdtest

from lettrade.exchange.backtest.data import CSVBackTestDataFeed


class DataFeedTestCase(unittest.TestCase):
    def setUp(self):
        self.data = CSVBackTestDataFeed("test/assets/EURUSD_1h-0_1000.csv")
        self.data._set_main()

        self.raw_data = pd.read_csv(
            "test/assets/EURUSD_1h-0_1000.csv",
            index_col=0,
            parse_dates=["datetime"],
        )

    def test_value(self):
        pdtest.assert_frame_equal(
            self.data,
            self.raw_data,
            "DataFrame is not equal pandas.DataFrame",
            check_index_type=False,
        )

    def test_next(self):
        df = self.data.copy(deep=True)

        for i in range(0, len(df)):
            row = df.l[0]
            raw_row = self.raw_data.iloc[i]

            pdtest.assert_series_equal(
                row,
                raw_row,
                f"DataFeed[{i}] is not equal",
                check_names=False,
            )

            self.assertEqual(row.open, raw_row.open, f"Data[{i}] open value wrong")
            self.assertEqual(row.high, raw_row.high, f"Data[{i}] high value wrong")
            self.assertEqual(row.low, raw_row.low, f"Data[{i}] low value wrong")
            self.assertEqual(row.close, raw_row.close, f"Data[{i}] close value wrong")
            self.assertEqual(
                row.volume, raw_row.volume, f"Data[{i}] volume value wrong"
            )

            df.next()

        self.assertEqual(self.data.l.pointer, 0, f"Data.pointer move after deepcopy")

    def test_iloc(self):
        df = self.data.copy(deep=True)

        for i in range(0, len(df)):
            row = df.iloc[i]
            raw_row = self.raw_data.iloc[i]

            pdtest.assert_series_equal(
                row,
                raw_row,
                f"DataFeed[{i}] is not equal",
                check_names=False,
            )

            self.assertEqual(row.open, raw_row.open, f"Data[{i}] open value wrong")
            self.assertEqual(row.high, raw_row.high, f"Data[{i}] high value wrong")
            self.assertEqual(row.low, raw_row.low, f"Data[{i}] low value wrong")
            self.assertEqual(row.close, raw_row.close, f"Data[{i}] close value wrong")
            self.assertEqual(
                row.volume, raw_row.volume, f"Data[{i}] volume value wrong"
            )

            df.next()

    def test_loc(self):
        df = self.data.copy(deep=True)

        self.assertEqual(df.l.pointer, 0, f"Data.pointer move after deepcopy")

        for i in range(0, len(df)):
            row = df.l[0]
            raw_row = self.raw_data.iloc[i]

            pdtest.assert_series_equal(
                row,
                raw_row,
                f"DataFeed[{i}] is not equal",
                check_names=False,
            )

            self.assertEqual(row.open, raw_row.open, f"Data[{i}] open value wrong")
            self.assertEqual(row.high, raw_row.high, f"Data[{i}] high value wrong")
            self.assertEqual(row.low, raw_row.low, f"Data[{i}] low value wrong")
            self.assertEqual(row.close, raw_row.close, f"Data[{i}] close value wrong")
            self.assertEqual(
                row.volume, raw_row.volume, f"Data[{i}] volume value wrong"
            )

            df.next()

    def test_shift(self):
        df = self.data.copy(deep=True)

        df_close1 = df.close.shift(1)
        self.assertEqual(
            df_close1.iloc[1], self.raw_data.close.iloc[0], f"Data.close.shift() wrong"
        )

        df_close3 = df_close1.shift(2)
        self.assertEqual(
            df_close3.iloc[3], self.raw_data.close.iloc[0], f"Data.close.shift() wrong"
        )

        self.assertIsInstance(
            df_close3.index,
            pd.DatetimeIndex,
            f"Data.close.index is not instance of pd.DatetimeIndex",
        )

        self.assertEqual(df.l.pointer, 0, f"Data.pointer move after deepcopy and shift")


if __name__ == "__main__":
    unittest.main(verbosity=2)



from .ichimoku import ichimoku
from .ma import ema, ma
from .parabolic_sar import parabolic_sar


def pandas_inject(obj: object | None = None):
    if obj is None:
        from pandas.core.base import PandasObject

        obj = PandasObject

    from ..series import series_indicator_inject

    obj.ma = series_indicator_inject(ma)
    obj.ema = series_indicator_inject(ema)
    obj.ichimoku = ichimoku
    obj.parabolic_sar = parabolic_sar



import unittest

import pandas as pd

from lettrade.data import DataFeed, TimeFrame
from lettrade.exchange.backtest.data import CSVBackTestDataFeed


class CSVBackTestDataFeedTestCase(unittest.TestCase):
    def setUp(self):
        self.data = CSVBackTestDataFeed("test/assets/EURUSD_1h-0_1000.csv")
        self.data._set_main()

    def test_types(self):
        self.assertIsInstance(
            self.data,
            DataFeed,
            "Data is not instance of DataFeed",
        )

        self.assertIsInstance(
            self.data.index,
            pd.DatetimeIndex,
            "Index is not instance of pd.DatetimeIndex",
        )

        # self.assertIsInstance(
        #     self.data.index,
        #     DataFeedIndex,
        #     "Index is not instance of DataFeedIndex",
        # )

        self.assertIsInstance(
            self.data.timeframe,
            TimeFrame,
            "timeframe is not instance of TimeFrame",
        )

    def test_name(self):
        self.assertEqual(self.data.name, "EURUSD_1h", "Name is wrong")

    def test_size(self):
        self.assertEqual(len(self.data), 1_000, "Size is wrong")

    def test_copy(self):
        df = self.data.copy(deep=True)
        # Test copy types
        self.assertIsInstance(
            df,
            self.data.__class__,
            f"Copy data is not instance of {self.data.__class__}",
        )
        self.assertIsInstance(
            df.index,
            self.data.index.__class__,
            f"Copy data.index is not instance of {self.data.index.__class__}",
        )
        self.assertIsInstance(
            df.timeframe,
            self.data.timeframe.__class__,
            f"Copy data.timeframe is not instance of {self.data.timeframe.__class__}",
        )

        # Test deepcopy is not a mirror
        loc = df.l.index[0]
        df.loc[loc, "open"] = 0
        self.assertEqual(df.loc[loc, "open"], 0, "Set value to data.open error")
        self.assertEqual(df.l[0].open, 0, "Set value to data.open error")
        self.assertNotEqual(self.data.loc[loc, "open"], 0, "Value change when deepcopy")

        loc = df.l.index[1]
        df.at[loc, "close"] = 1
        self.assertEqual(df.loc[loc, "close"], 1, "Set value to data.open error")
        self.assertEqual(df.l[1].close, 1, "Set value to data.open error")
        self.assertNotEqual(
            self.data.loc[loc, "close"], 1, "Value change when deepcopy"
        )

    # Test drop
    def test_drop(self):
        df: DataFeed = self.data.copy(deep=True)
        df.drop(since=100)

        self.assertEqual(len(df), 900, "Drop data size wrong")
        self.assertEqual(df.l.open[0], 0.99474, "Drop data open value wrong")
        self.assertEqual(len(self.data), 1_000, "self.data size wrong")
        self.assertEqual(self.data.l.open[0], 0.97724, "self.data open value wrong")

    # Test index
    def test_index(self):
        df = self.data.copy(deep=True)

        # Move to nexts rows
        next = 3
        df.next(next)
        self.assertEqual(df.l.pointer, next, "Data pointer wrong")

        self.assertEqual(df.l.open[0], 0.97656, f"Data.open[{next}] wrong")
        self.assertEqual(df.l.high[0], 0.97718, f"Data.high[{next}] wrong")
        self.assertEqual(df.l.low[0], 0.97585, f"Data.low[{next}] wrong")
        self.assertEqual(df.l.close[0], 0.9765, f"Data.close[{next}] wrong")
        self.assertEqual(df.l.volume[0], 5050.0, f"Data.volume[{next}] wrong")
        self.assertEqual(
            df.l.index[0],
            pd.Timestamp("2022-10-20 03:00:00"),
            f"Data.index[{next}] wrong",
        )

        # Move to end
        end = len(df) - 1
        df.l.go_stop()
        self.assertEqual(df.l.pointer, end, "Data pointer wrong")
        self.assertEqual(df.l.open[0], 1.06215, f"Data.open[{end}] wrong")
        self.assertEqual(df.l.high[0], 1.06359, f"Data.high[{end}] wrong")
        self.assertEqual(df.l.low[0], 1.06134, f"Data.low[{end}] wrong")
        self.assertEqual(df.l.close[0], 1.06341, f"Data.close[{end}] wrong")
        self.assertEqual(df.l.volume[0], 5679.0, f"Data.volume[{end}] wrong")
        self.assertEqual(
            df.l.index[0],
            pd.Timestamp("2022-12-16 15:00:00"),
            f"Data.index[{end}] wrong",
        )


if __name__ == "__main__":
    unittest.main(verbosity=2)



from lettrade.base.error import LetException


class LetOrderException(LetException):
    """Base order exception"""


class LetOrderValidateException(LetOrderException):
    """Order attribute validate exception"""


class LetOrderInvalidException(LetOrderException):
    """Order place is invalid exception"""



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from lettrade.account import Account
from lettrade.data import DataFeeder

if TYPE_CHECKING:
    # Recusive
    from lettrade import LetTradeBot
    from lettrade.brain import Brain
    from lettrade.exchange import Exchange
    from lettrade.plot import Plotter
    from lettrade.strategy import Strategy


class Commander(ABC):
    """
    Abstract class for strategy commander. Help to manage and report strategy real-time
    """

    bot: "LetTradeBot"
    brain: "Brain"
    feeder: DataFeeder
    exchange: "Exchange"
    account: Account
    strategy: "Strategy"
    plotter: "Plotter | None" = None

    _name: str

    def init(
        self,
        bot: "LetTradeBot",
        brain: "Brain",
        exchange: "Exchange",
        strategy: "Strategy",
    ):
        """Init commander dependencies

        Args:
            bot (LetTradeBot): LetTradeBot object
            brain (Brain): Brain of bot
            exchange (Exchange): Manage bot trading
            strategy (Strategy): Strategy of bot
        """
        self.bot = bot
        self.brain = brain
        self.exchange = exchange
        self.strategy = strategy

        self._name = self.bot._name

    @abstractmethod
    def start(self):
        pass

    @abstractmethod
    def stop(self):
        pass

    @abstractmethod
    def send_message(self, msg: str, **kwargs):
        pass

    @classmethod
    def multiprocess(cls, kwargs, **other_kwargs):
        pass
